/****************************************************************************
** $Id: moc.y,v 1.1 2001/05/29 21:43:57 mduncan Exp $
**
** Parser and code generator for meta object compiler
**
** Created : 930417
**
** Copyright (C) 1993-1999 by Troll Tech AS.  All rights reserved.
**
** This file is part of the Qt GUI Toolkit.
**
** This file may be distributed under the terms of the Q Public License
** as defined by Troll Tech AS of Norway and appearing in the file
** LICENSE.QPL included in the packaging of this file.
**
** Licensees holding valid Qt Professional Edition licenses may use this
** file in accordance with the Qt Professional Edition License Agreement
** provided with the Qt Professional Edition.
**
** See http://www.troll.no/pricing.html or email sales@troll.no for
** information about the Professional Edition licensing, or see
** http://www.troll.no/qpl/ for QPL licensing information.
**
** --------------------------------------------------------------------------
**
** This compiler reads a C++ header file with class definitions and ouputs
** C++ code to build a meta class. The meta data includes public methods
** (not constructors, destructors or operator functions), signals and slot
** definitions. The output file should be compiled and linked into the
** target application.
**
** C++ header files are assumed to have correct syntax, and we are therefore
** doing less strict checking than C++ compilers.
**
** The C++ grammar has been adopted from the "The Annotated C++ Reference
** Manual" (ARM), by Ellis and Stroustrup (Addison Wesley, 1992).
**
** Notice that this code is not possible to compile with GNU bison, instead
** use standard AT&T yacc or Berkeley yacc.
*****************************************************************************/

%{
void yyerror( char *msg );

#include "qlist.h"
#include "qdict.h"
#include "qstring.h"
#include "qdatetime.h"
#include "qregexp.h"
#include "qasciidict.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>


static QCString rmWS( const char * );

enum AccessPerm { _PRIVATE, _PROTECTED, _PUBLIC };


struct Argument					// single arg meta data
{
    Argument( char *left, char *right, char *name = "" )
	{ leftType  = rmWS( left );
	  rightType = rmWS( right );
	  if ( leftType == "void" && rightType.isEmpty() )
	      leftType = "";
	  this->name = rmWS( name );
	}
    QCString leftType;
    QCString rightType;
    QCString name;
};

class ArgList : public QList<Argument> {	// member function arg list
public:
    ArgList() { setAutoDelete(TRUE); }
   ~ArgList() { clear(); }
};

struct EnumValue  {
  EnumValue( const char *name, int value ) {
    this->name = name;
    this->value = value;
  }

  int		value;
  QCString	name;
};

class EnumDef : public QList<EnumValue> {
 public:
  EnumDef(char *name)	{ setAutoDelete(TRUE); this->name = name; assigned = FALSE; }
  ~EnumDef()		{ clear(); }

  QCString	name;
  bool          assigned;
};

class EnumDefs : public QList<EnumDef> {
 public:
  EnumDefs()		{ setAutoDelete(TRUE); }
  ~EnumDefs()		{ clear(); }
};

struct Function					// member function meta data
{						//   used for signals and slots
    AccessPerm accessPerm;
    QCString    qualifier;			// const or volatile
    QCString    name;
    QCString    type;
    int	       lineNo;
    bool	is_static;
    ArgList   *args;
    Function() { args=0; }
   ~Function() { delete args; }
};

class FuncList : public QList<Function> {	// list of member functions
public:
    FuncList() { setAutoDelete(TRUE); }
};


ArgList *addArg( Argument * );			// add arg to tmpArgList
void	 addMember( char );			// add tmpFunc to current class

char	*strnew( const char * );		// returns a new string (copy)
char	*stradd( const char *, const char * );	// add two strings
char	*stradd( const char *, const char *,	// add three strings
			       const char * );
char	*straddSpc( const char *, const char * );
char	*straddSpc( const char *, const char *,
			       const char * );
char	*straddSpc( const char *, const char *,
		    const char *, const char * );

extern int yydebug;
bool	   lexDebug	   = FALSE;
bool	   grammarDebug	   = FALSE;

bool       binding         = FALSE;
bool	   cHeader	   = FALSE;
bool	   cImpl	   = FALSE;
bool	   cTypes	   = FALSE;
bool       cHook           = FALSE;
bool	   pascalBinding   = FALSE;
bool	   pascalIntf	   = FALSE;
bool	   pascalImpl	   = FALSE;
bool       pascalTypes     = FALSE;
bool       pascalEnums     = FALSE;
bool       pascalEvents    = FALSE;
bool       vcl             = FALSE;
bool       main_gened      = FALSE;


int	   lineNo;				// current line number
bool	   errorControl	   = FALSE;		// controlled errors
bool	   displayWarnings = TRUE;
bool	   skipClass;				// don't generate for class
bool	   skipFunc;				// don't generate for func
bool	   staticFunc;				// function is a static class func
bool	   templateClass;			// class is a template

ArgList	  *tmpArgList;				// current argument list
Function  *tmpFunc;				// current member function
AccessPerm tmpAccessPerm;			// current access permission
AccessPerm subClassPerm;			// current access permission
bool	   Q_OBJECTdetected;			// TRUE if current class
						// contains the Q_OBJECT macro
QCString   tmpExpression;

const int  formatRevision = 3;			// moc output format revision

bool	   assignedEnum;
EnumDef	   *tmpEnum;
EnumDefs   enumDefs;
int	   cur_enum;
%}


%union {
    char	char_val;
    int		int_val;
    double	double_val;
    char       *string;
    AccessPerm	access;
    Function   *function;
    ArgList    *arg_list;
    Argument   *arg;
    EnumDef    *enum_def;
    EnumValue  *enum_value;
}

%start declaration_seq

%token <char_val>	CHAR_VAL		/* value types */
%token <int_val>	INT_VAL
%token <double_val>	DOUBLE_VAL
%token <string>		STRING
%token <string>		IDENTIFIER		/* identifier string */

%token			FRIEND			/* declaration keywords */
%token			TYPEDEF
%token			AUTO
%token			REGISTER
%token			STATIC
%token			EXTERN
%token			INLINE
%token			VIRTUAL
%token			CONST
%token			VOLATILE
%token			CHAR
%token			SHORT
%token			INT
%token			LONG
%token			SIGNED
%token			UNSIGNED
%token			FLOAT
%token			DOUBLE
%token			VOID
%token			ENUM
%token			CLASS
%token			STRUCT
%token			UNION
%token			ASM
%token			PRIVATE
%token			PROTECTED
%token			PUBLIC
%token			OPERATOR
%token			DBL_COLON
%token			TRIPLE_DOT
%token			TEMPLATE
%token			NAMESPACE
%token			USING
%token			MUTABLE

%token			SIGNALS
%token			SLOTS
%token			Q_OBJECT
%token			QT_EXPORT

%type  <string>		class_name
%type  <string>		template_class_name
%type  <string>		template_spec
%type  <string>		opt_base_spec

%type  <string>		base_spec
%type  <string>		base_list
%type  <string>		qt_macro_name
%type  <string>		base_specifier
%type  <access>		access_specifier
%type  <string>		fct_name
%type  <string>		type_name
%type  <string>		simple_type_names
%type  <string>		simple_type_name
%type  <string>		class_key
%type  <string>		complete_class_name
%type  <string>		qualified_class_name
%type  <string>		elaborated_type_specifier

%type  <arg_list>	argument_declaration_list
%type  <arg_list>	arg_declaration_list
%type  <arg_list>	arg_declaration_list_opt
%type  <string>		abstract_decl_opt
%type  <string>		abstract_decl
%type  <arg>		argument_declaration
%type  <string>		cv_qualifier_list_opt
%type  <string>		cv_qualifier_list
%type  <string>		cv_qualifier
%type  <string>		decl_specifiers
%type  <string>		decl_specifier
%type  <string>		decl_specs_opt
%type  <string>		decl_specs
%type  <string>		type_specifier
%type  <string>		declarator
%type  <string>		dname
%type  <string>		ptr_operator
%type  <string>		ptr_operators
%type  <string>		ptr_operators_opt

%type  <enum_def>	enum_specifier
%type  <enum_def>	enum_list_opt
%type  <enum_def>	enum_list
%type  <enum_value>	enumerator

%%
declaration_seq:	  /* empty */
                        | declaration_seq declaration
			;

declaration:		  class_def
/* | template_declaration */
                        | namespace_def
                        | namespace_alias_def
			| using_declaration
                        | using_directive
                        ;

namespace_def:            named_namespace_def
                        | unnamed_namespace_def
                        ;

named_namespace_def:      NAMESPACE
                          IDENTIFIER         { enterNameSpace($2); }
                          '{'                { BEGIN IN_NAMESPACE; }
                          namespace_body
                          '}'                { leaveNameSpace();
			                       selectOutsideClassState();
                                             }
                        ;

unnamed_namespace_def:    NAMESPACE          { enterNameSpace(); }
			  '{'                { BEGIN IN_NAMESPACE; }
                          namespace_body
                          '}'                { leaveNameSpace();
  			                       selectOutsideClassState();
			                     }
                        ;

namespace_body:           declaration_seq
                        ;

namespace_alias_def:      NAMESPACE IDENTIFIER '=' complete_class_name ';'
                                            { selectOutsideClassState(); }
                        ;


using_directive:          USING NAMESPACE   { selectOutsideClassState(); } /* Skip namespace */
                        ;

using_declaration:        USING IDENTIFIER  { selectOutsideClassState(); }
                        | USING DBL_COLON   { selectOutsideClassState(); }
                        ;

class_def:				      { initClass(); }
			  class_specifier ';' {
						if ( pascalBinding )
						  generatePascal();
						else if ( pascalEnums )
						  generateEnumValues();
						else if ( binding )
						  generateCBinding();
						else
						  generateClass();
						registerClassInNamespace();
						selectOutsideClassState(); }
			;


/***** r.17.1 (ARM p.387 ): Keywords	*****/

class_name:		  IDENTIFIER	      { $$ = $1; }
			| template_class_name { $$ = $1; }
			;

template_class_name:	  IDENTIFIER '<' template_args '>'
				   { $$ = stradd( $1, "<",
				     tmpExpression =
				     tmpExpression.stripWhiteSpace(), ">" ); }
			;

/*
   template_args skips all characters until it encounters a ">" (it
   handles and discards sublevels of parentheses).  Since the rule is
   empty it must be used with care!
*/

template_args:		  /* empty */		  { initExpression();
						    templLevel = 1;
						    BEGIN IN_TEMPL_ARGS; }
			;

/***** r.17.2 (ARM p.388): Expressions	*****/


/* const_expression skips all characters until it encounters either one
   of "]", ")" or "," (it handles and discards sublevels of parentheses).
   Since the rule is empty it must be used with care!
*/

const_expression:	   /* empty */		  { initExpression();
						    BEGIN IN_EXPR; }
			;

/***** r.17.3 (ARM p.391): Declarations *****/

decl_specifier:		  storage_class_specifier { $$ = ""; }
			| type_specifier	  { $$ = $1; }
			| fct_specifier		  { $$ = ""; }
			| FRIEND		  { skipFunc = TRUE; $$ = ""; }
			| TYPEDEF		  { skipFunc = TRUE; $$ = ""; }
			;

decl_specifiers:	  decl_specs_opt type_name decl_specs_opt
						  { $$ = straddSpc($1,$2,$3); }
			;

decl_specs_opt:			/* empty */	  { $$ = ""; }
			| decl_specs		  { $$ = $1; }
			;

decl_specs:		  decl_specifier	    { $$ = $1; }
			| decl_specs decl_specifier { $$ = straddSpc($1,$2); }
			;

storage_class_specifier:  AUTO
			| REGISTER
			| STATIC		{ if ( !binding ) skipFunc = TRUE;
						  staticFunc = TRUE; }
			| EXTERN
			| QT_EXPORT
			;

fct_specifier:		  INLINE
			| VIRTUAL
			;

type_specifier:		  CONST			{ $$ = "const"; }
			| VOLATILE		{ $$ = "volatile"; }
			;

type_name:		  elaborated_type_specifier { $$ = $1; }
			| complete_class_name	    { $$ = $1; }
			| simple_type_names	    { $$ = $1; }
			;

simple_type_names:	  simple_type_names simple_type_name
						    { $$ = straddSpc($1,$2); }
			| simple_type_name	    { $$ = $1; }

simple_type_name:	  CHAR			    { $$ = "char"; }
			| SHORT			    { $$ = "short"; }
			| INT			    { $$ = "int"; }
			| LONG			    { $$ = "long"; }
			| SIGNED		    { $$ = "signed"; }
			| UNSIGNED		    { $$ = "unsigned"; }
			| FLOAT			    { $$ = "float"; }
			| DOUBLE		    { $$ = "double"; }
			| VOID			    { $$ = "void"; }
			;

template_spec:		  TEMPLATE '<' template_args '>'
				   { $$ = stradd( "template<",
				     tmpExpression =
				     tmpExpression.stripWhiteSpace(), ">" ); }
			;

opt_template_spec:	  /* empty */
			| template_spec		{ templateClass = TRUE; }
			;


class_key:		  opt_template_spec CLASS { $$ = "class"; }
			| opt_template_spec STRUCT { $$ = "struct"; }
			;

complete_class_name:	  qualified_class_name	{ $$ = $1; }
			| DBL_COLON  qualified_class_name
						{ $$ = stradd( "::", $2 ); }
			;

qualified_class_name:	  qualified_class_name DBL_COLON class_name
						{ $$ = stradd( $1, "::", $3 );}
			| class_name		{ $$ = $1; }
			;

elaborated_type_specifier:
			  class_key IDENTIFIER	{ $$ = straddSpc($1,$2); }
			| ENUM IDENTIFIER	{ $$ = stradd("enum ",$2); }
			| UNION IDENTIFIER	{ $$ = stradd("union ",$2); }
			;

/***** r.17.4 (ARM p.393): Declarators	*****/

argument_declaration_list:  arg_declaration_list_opt triple_dot_opt { $$ = $1;}
			|   arg_declaration_list ',' TRIPLE_DOT	    { $$ = $1;
			        if ( binding )
				  moc_warn( "Ellipsis not supported in "
					    "C binding.\n"
					    "Ellipsis argument ignored.");
				else
				  moc_warn("Ellipsis not supported"
					   " in signals and slots.\n"
					   "Ellipsis argument ignored."); }
			;

arg_declaration_list_opt:	/* empty */	{ $$ = tmpArgList; }
			| arg_declaration_list	{ $$ = $1; }
			;

triple_dot_opt:			/* empty */
			| TRIPLE_DOT {
			  if ( binding )
			    moc_warn("Ellipsis not supported "
				     "in C binding.\n"
				     "Ellipsis argment ignored.");
			  else
			    moc_warn("Ellipsis not supported"
				     " in signals and slots.\n"
				     "Ellipsis argument ignored."); }

			;

arg_declaration_list:	  arg_declaration_list
			  ','
			  argument_declaration	{ $$ = addArg($3); }
			| argument_declaration	{ $$ = addArg($1); }

argument_declaration:	  decl_specifiers abstract_decl_opt
				{ $$ = new Argument(straddSpc($1,$2),"");
				  CHECK_PTR($$); }
			| decl_specifiers abstract_decl_opt
			  '=' { expLevel = 1; }
			  const_expression
				{ $$ = new Argument(straddSpc($1,$2),"");
				  CHECK_PTR($$); }
			| decl_specifiers abstract_decl_opt dname
				abstract_decl_opt
				{ $$ = new Argument(straddSpc($1,$2),$4,$3);
				  CHECK_PTR($$); }
			| decl_specifiers abstract_decl_opt dname
				abstract_decl_opt
			  '='	{ expLevel = 1;
			}
			  const_expression
				{ $$ = new Argument(straddSpc($1,$2),$4,$3);
				  CHECK_PTR($$); }
			;


abstract_decl_opt:	  /* empty */		{ $$ = ""; }
			| abstract_decl		{ $$ = $1; }
			;

abstract_decl:		  abstract_decl ptr_operator
						{ $$ = straddSpc($1,$2); }
			| '['			{ expLevel = 1; }
			  const_expression ']'
				   { $$ = stradd( "[",
				     tmpExpression =
				     tmpExpression.stripWhiteSpace(), "]" ); }
			| abstract_decl '['	{ expLevel = 1; }
			  const_expression ']'
				   { $$ = stradd( $1,"[",
				     tmpExpression =
				     tmpExpression.stripWhiteSpace(),"]" ); }
			| ptr_operator		{ $$ = $1; }
			| '(' abstract_decl ')' { $$ = $2; }
			;

declarator:		  dname			{ $$ = ""; }
			| declarator ptr_operator
						{ $$ = straddSpc($1,$2);}
			| declarator '['	{ expLevel = 1; }
			  const_expression ']'
				   { $$ = stradd( $1,"[",
				     tmpExpression =
				     tmpExpression.stripWhiteSpace(),"]" ); }
			| '(' declarator ')'	{ $$ = $2; }
			;

dname:			  IDENTIFIER		{ $$ = $1; }
			;

fct_decl:		  '('
			  argument_declaration_list
			  ')'
			  cv_qualifier_list_opt
			  ctor_initializer_opt
			  fct_body_or_semicolon
						{ tmpFunc->args	     = $2;
						  tmpFunc->qualifier = $4; }
			;

fct_name:		  IDENTIFIER		/* NOTE: simplified! */
			| IDENTIFIER array_decls
				{ func_warn("Variable as signal or slot."); }
			| IDENTIFIER '=' { expLevel=0; }
			  const_expression     /* probably const member */
						{ skipFunc = TRUE; }
			| IDENTIFIER array_decls '=' { expLevel=0; }
			  const_expression     /* probably const member */
						{ skipFunc = TRUE; }
			;


array_decls:		 '['			{ expLevel = 1; }
			  const_expression ']'
			| array_decls '['	 { expLevel = 1; }
			  const_expression ']'

			;

ptr_operators_opt:	   /* empty */		{ $$ = ""; }
			| ptr_operators		 { $$ = $1; }
			;

ptr_operators:		  ptr_operator		{ $$ = $1; }
			| ptr_operators ptr_operator { $$ = straddSpc($1,$2);}
			;

ptr_operator:		  '*' cv_qualifier_list_opt { $$ = straddSpc("*",$2);}
			| '&' cv_qualifier_list_opt { $$ = stradd("&",$2);}
/*!			| complete_class_name
			  DBL_COLON
			  '*'
			  cv_qualifier_list_opt { $$ = stradd($1,"::*",$4); }*/
			;

cv_qualifier_list_opt:		/* empty */	{ $$ = ""; }
			| cv_qualifier_list	{ $$ = $1; }
			;

cv_qualifier_list:	  cv_qualifier		{ $$ = $1; }
			| cv_qualifier_list cv_qualifier
						{ $$ = straddSpc($1,$2); }
			;

cv_qualifier:		  CONST			{ $$ = "const"; }
			| VOLATILE		{ $$ = "volatile"; }
			;

fct_body_or_semicolon:	  ';'
			| fct_body
			| '=' INT_VAL ';'   /* abstract func, INT_VAL = 0 */
			;

fct_body:		  '{' {BEGIN IN_FCT; fctLevel = 1;}
			  '}' {BEGIN QT_DEF; }
			;


/***** r.17.5 (ARM p.395): Class Declarations *****/

class_specifier:	  full_class_head
			  '{'			{ BEGIN IN_CLASS;
                                                  classPLevel = 1; }
			  opt_obj_member_list
			  '}'			{ BEGIN QT_DEF; } /*catch ';'*/
			| class_head		{ BEGIN QT_DEF;	  /* -- " -- */
						  skipClass = TRUE; }
			| class_head
			  '(' IDENTIFIER ')' /* Qt macro name */
						{ BEGIN QT_DEF; /* catch ';' */
						  skipClass = TRUE; }
			| template_spec whatever { skipClass = TRUE;
						  BEGIN GIMME_SEMICOLON; }
			;

whatever:		  IDENTIFIER
			| simple_type_name
			| type_specifier
			| storage_class_specifier
			| fct_specifier
			;


class_head:		  class_key
			  qualified_class_name	{ className = $2;  }
			| class_key
			  export_decl		/* possible DLL EXPORT macro */
			  class_name		{ className = $3; }
			;

export_decl:		  IDENTIFIER
			| QT_EXPORT
			;

full_class_head:	  class_head
			  opt_base_spec		{ superclassName = $2; }
			;

opt_base_spec:			/* empty */	{ $$ = 0; }
			| base_spec		{ $$ = $1; }
			;

opt_obj_member_list:		/* empty */
			| obj_member_list
			;

obj_member_list:	  obj_member_list obj_member_area
			| obj_member_area
			;


qt_access_specifier:	  access_specifier	{ tmpAccessPerm = $1; }
			| SLOTS	      { moc_err( "Missing access specifier"
						   " before \"slots:\"." ); }
			;

obj_member_area:	  qt_access_specifier	{ BEGIN QT_DEF; }
			  slot_area
			| SIGNALS		{ BEGIN QT_DEF; tmpAccessPerm = _PUBLIC; }
			  ':'  opt_signal_declarations
			| Q_OBJECT		{ if ( tmpAccessPerm )
				moc_warn("Q_OBJECT is not in the private"
					" section of the class.\n"
					"Q_OBJECT is a macro that resets"
					" access permission to \"private\".");
						  Q_OBJECTdetected = TRUE; }

			;

slot_area:		  SIGNALS ':'	     { moc_err( "Signals cannot "
						 "have access specifiers" ); }
			| SLOTS	  ':' opt_slot_declarations
			| ':'		 { if ( grammarDebug ||
						(binding &&
						 tmpAccessPerm == _PUBLIC) )
						  BEGIN QT_DEF;
					      else
						  BEGIN IN_CLASS;
					 }
				      opt_slot_declarations
			| IDENTIFIER	 { BEGIN IN_CLASS;
					   if ( classPLevel != 1 )
					       moc_warn( "unexpected access"
							 "specifier" );
					 }
			;

opt_signal_declarations:	/* empty */
			| signal_declarations
			;

signal_declarations:	  signal_declarations signal_declaration
			| signal_declaration
			;


signal_declaration:	  signal_or_slot	{ addMember('s'); }
			;

opt_slot_declarations:		/* empty */
			| slot_declarations
			;

slot_declarations:	  slot_declarations slot_declaration
			| slot_declaration
			;

slot_declaration:	  signal_or_slot	{ addMember('t'); }
			;

opt_semicolons:		  /* empty */
			| opt_semicolons ';'
			;

base_spec:		  ':' { superclassList.clear(); } base_list { $$ = $3; }
			;

base_list		: base_list ',' base_specifier  { addSuperClass($3); $$ = $3; }
			| base_specifier                { addSuperClass($1); $$ = $1; }
			;

qt_macro_name:		  IDENTIFIER '(' IDENTIFIER ')'
					   { $$ = stradd( $1, "(", $3, ")" ); }
			| IDENTIFIER '(' simple_type_name ')'
					   { $$ = stradd( $1, "(", $3, ")" ); }
			;

base_specifier:		  complete_class_name			       {$$=$1;}
			| VIRTUAL access_specifier complete_class_name {$$=$3;}
			| VIRTUAL complete_class_name		       {$$=$2;}
			| access_specifier VIRTUAL complete_class_name {$$=$3;}
			| access_specifier complete_class_name	       {$$=$2;}
			| qt_macro_name				       {$$=$1;}
			| VIRTUAL access_specifier qt_macro_name       {$$=$3;}
			| VIRTUAL qt_macro_name			       {$$=$2;}
			| access_specifier VIRTUAL qt_macro_name       {$$=$3;}
			| access_specifier qt_macro_name	       {$$=$2;}
			;

access_specifier:	  PRIVATE		{ $$=_PRIVATE; }
			| PROTECTED		{ $$=_PROTECTED; }
			| PUBLIC		{ $$=_PUBLIC; }
			;

operator_name:		  decl_specs_opt IDENTIFIER ptr_operators_opt
			| decl_specs_opt simple_type_name ptr_operators_opt
			| '+'
			| '-'
			| '*'
			| '/'
			| '%'
			| '^'
			| '&'
			| '|'
			| '~'
			| '!'
			| '='
			| '<'
			| '>'
			| '+' '='
			| '-' '='
			| '*' '='
			| '/' '='
			| '%' '='
			| '^' '='
			| '&' '='
			| '|' '='
			| '~' '='
			| '!' '='
			| '=' '='
			| '<' '='
			| '>' '='
			| '<' '<'
			| '>' '>'
			| '<' '<' '='
			| '>' '>' '='
			| '&' '&'
			| '|' '|'
			| '+' '+'
			| '-' '-'
			| ','
			| '-' '>' '*'
			| '-' '>'
			| '(' ')'
			| '[' ']'
			;


opt_virtual:		  /* empty */
			| VIRTUAL
			;

type_and_name:		  type_name fct_name
						{ tmpFunc->type = $1;
						  tmpFunc->name = $2; }
			| fct_name
						{ tmpFunc->type = "int";
						  tmpFunc->name = $1;
				  if ( tmpFunc->name == className )
				    if ( binding )
				      tmpFunc->type = className;
				    else
				      func_warn( "Constructors cannot be"
						 " signals or slots.");
						}
			| opt_virtual '~' fct_name
						{
				     if ( binding )
				       skipFunc = TRUE;
				     else
				       func_warn( "Destructors cannot be"
						  " signals or slots.");
						}
			| decl_specs type_name decl_specs_opt
			  ptr_operators_opt fct_name
						{
						    char *tmp =
							straddSpc($1,$2,$3,$4);
						    tmpFunc->type = rmWS(tmp);
						    delete tmp;
						    tmpFunc->name = $5; }
			| decl_specs type_name /* probably friend decl */
						{ skipFunc = TRUE; }
			| type_name ptr_operators fct_name
						{ tmpFunc->type =
						      straddSpc($1,$2);
						  tmpFunc->name = $3; }
			| type_name decl_specs ptr_operators_opt
			  fct_name
						{ tmpFunc->type =
						      straddSpc($1,$2,$3);
						  tmpFunc->name = $4; }
			| type_name OPERATOR operator_name
						{ operatorError();    }
			| OPERATOR operator_name
						{ operatorError();    }
			| decl_specs type_name decl_specs_opt
			  ptr_operators_opt OPERATOR operator_name
						{ operatorError();    }
			| type_name ptr_operators OPERATOR  operator_name
						{ operatorError();    }
			| type_name decl_specs ptr_operators_opt
			  OPERATOR  operator_name
						{ operatorError();    }
			;

signal_or_slot:		  type_and_name fct_decl ctor_initializer_opt opt_semicolons
			| type_and_name opt_bitfield ';' opt_semicolons
				{ func_warn("Variable as signal or slot."); }
			| type_and_name opt_bitfield ','member_declarator_list
			  ';' opt_semicolons
				{
				  if ( binding )
				    skipFunc = TRUE;
				  else
				    func_warn("Variable as signal or slot.");
				}
			| enum_specifier  ';' opt_semicolons
				{
				  if ( binding ) {
				    addEnumDef($1);
				    skipFunc = TRUE;
				  }
				  else
				    func_warn("Enum declaration as signal or"
					  " slot.");
				}
                        | USING complete_class_name ';' opt_semicolons
                                { func_warn("Using declaration as signal or"
					    " slot."); }
			|USING NAMESPACE complete_class_name ';' opt_semicolons
                                { func_warn("Using declaration as signal or"
					    " slot."); }
			| NAMESPACE IDENTIFIER '{'
                                { classPLevel++;
				  moc_err("Namespace declaration as signal or"
					  " slot."); }
			;


member_declarator_list:	  member_declarator
			| member_declarator_list ',' member_declarator
			;

member_declarator:	  declarator
			| IDENTIFIER ':'	 { expLevel = 0; }
			  const_expression
			| ':'			 { expLevel = 0; }
			  const_expression
			;

opt_bitfield:		  /* empty */
			| ':'			 { expLevel = 0; }
			  const_expression
			;

enum_specifier:		  ENUM IDENTIFIER
				{
				  cur_enum = 0;
				  assignedEnum = FALSE;
				  tmpEnum = new EnumDef($2);
				}
			   '{' enum_list_opt '}'
				{ $$ = tmpEnum; tmpEnum->assigned = assignedEnum; }
			| ENUM { tmpEnum = NULL; } '{' enum_list_opt '}'
				{ $$ = NULL; }
			;

enum_list_opt:					     { $$ = NULL; }
			| enum_list		     { $$ = $1; }
			;

enum_list:		  enumerator                 { addEnum($1); }
			| enum_list ',' enumerator   { addEnum($3); }
			;

enumerator:		  IDENTIFIER { $$ = new EnumValue($1, cur_enum++); }
			| IDENTIFIER { expLevel = 0; } '='  const_expression
				{
				  $$ = new EnumValue($1, -1);
                                  assignedEnum = TRUE;
				}
			;

ctor_initializer_opt:	  /* empty */
			| ctor_initializer
			;

ctor_initializer:	  ':' mem_initializer_list
			;

mem_initializer_list:	  mem_initializer
			| mem_initializer_list ',' mem_initializer
			;

mem_initializer:	  complete_class_name '(' const_expr_list ')'
			| IDENTIFIER '(' const_expr_list ')'
			;

const_expr_list:	  const_expr_list ',' { expLevel = 1; } const_expression
			| { expLevel = 1; } const_expression
			;

/*
enum_specifier:		  ENUM opt_identifier '{'   {BEGIN IN_FCT; fctLevel=1;}
					      '}'   {BEGIN QT_DEF; }
			;
*/


/*opt_identifier:		  /* empty */
/*			| IDENTIFIER
			;
*/
%%

#if defined(_OS_WIN32_)
#include <io.h>
#undef isatty
extern "C" int hack_isatty( int )
{
    return 0;
}
#define isatty hack_isatty
#else
#include <unistd.h>
#endif

#include "lex.yy.c"

void 	  init();				// initialize
void 	  initClass();				// prepare for new class
void	  checkClassDup();			// Make sure class is not a duplicate
void 	  generateClass();			// generate C++ code for class
void 	  initExpression();			// prepare for new expression
QCString  combinePath( const char *, const char * );

QCString  fileName;				// file name
QCString  outputFile;				// output file name
QCString  includeFile;				// name of #include file
QCString  includePath;				// #include file path
QCString  qtPath;				// #include qt file path
bool	  noInclude     = FALSE;		// no #include <filename>
bool	  generatedCode = FALSE;		// no code generated
bool	  mocError = FALSE;			// moc parsing error occurred
QCString  className;				// name of parsed class
QCString  superclassName;			// name of super class
QList<QCString> superclassList;			// list of all super classes
QList<QCString> classList;			// list of all classes generated
FuncList  signals;				// signal interface
FuncList  slots;				// slots interface

FILE  *out;					// output file

int yyparse();

void replace( char *s, char c1, char c2 );

int main( int argc, char **argv )
{
    bool autoInclude = TRUE;
    char *error	     = 0;
    qtPath = "";
    for ( int n=1; n<argc && error==0; n++ ) {
	QCString arg = argv[n];
	if ( arg[0] == '-' ) {			// option
	    QCString opt = &arg[1];
	    if ( opt[0] == 'o' ) {		// output redirection
		if ( opt[1] == '\0' ) {
		    if ( !(n < argc-1) ) {
			error = "Missing output-file name";
			break;
		    }
		    outputFile = argv[++n];
		} else
		    outputFile = &opt[1];
	    } else if ( opt == "i" ) {		// no #include statement
		noInclude   = TRUE;
		autoInclude = FALSE;
	    } else if ( opt[0] == 'f' ) {	// produce #include statement
		noInclude   = FALSE;
		autoInclude = FALSE;
		if ( opt[1] ) {			// -fsomething.h
		    includeFile = &opt[1];
		}
	    } else if ( opt[0] == 'p' ) {	// include file path
		if ( opt[1] == '\0' ) {
		    if ( !(n < argc-1) ) {
			error = "Missing path name for the -p option.";
			break;
		    }
		    includePath = argv[++n];
		} else {
		    includePath = &opt[1];
		}
	    } else if ( opt[0] == 'q' ) {	// qt include file path
		if ( opt[1] == '\0' ) {
		    if ( !(n < argc-1) ) {
			error = "Missing path name for the -q option.";
			break;
		    }
		    qtPath = argv[++n];
		} else {
		    qtPath = &opt[1];
		}
		replace(qtPath.data(),'\\','/');
		if ( qtPath.right(1) != "/" )
		    qtPath += '/';
	    } else if ( opt == "k" ) {		// don't stop on errors
		errorControl = TRUE;
	    } else if ( opt == "nw" ) {		// don't display warnings
		displayWarnings = FALSE;
	    } else if ( opt == "ldbg" ) {	// lex debug output
		lexDebug = TRUE;
	    } else if ( opt == "ydbg" ) {	// yacc debug output
		yydebug = TRUE;
	    } else if ( opt == "dbg" ) {	// non-signal members are slots
		grammarDebug = TRUE;
	    } else if ( opt == "cheader" ) {
	        binding = TRUE;
		cHeader = TRUE;
	    } else if ( opt == "cimpl" ) {
		binding = TRUE;
		cImpl = TRUE;
	    } else if ( opt == "ctypes" ) {
		binding = TRUE;
		cTypes = TRUE;
            } else if ( opt == "chook" ) {
                binding = TRUE;
                cHook = TRUE;
	    } else if ( opt == "PascalIntf" ) {
	        binding = TRUE;
		pascalBinding = TRUE;
		pascalIntf = TRUE;
	    } else if ( opt == "PascalImpl" ) {
		binding = TRUE;
		pascalBinding = TRUE;
		pascalImpl = TRUE;
	    } else if ( opt == "PascalTypes" ) {
		binding = TRUE;
		pascalBinding = TRUE;
		pascalTypes = TRUE;
            } else if ( opt == "PascalEnums" ) {
                binding = TRUE;
		pascalEnums = TRUE;
            } else if ( opt == "PascalEvents" ) {
                binding = TRUE;
                pascalBinding = TRUE;
                pascalEvents = TRUE;
	    } else {
		error = "Invalid argument";
	    }
	} else {
	    if ( !fileName.isNull() )		// can handle only one file
		error	 = "Too many input files specified";
	    else
		fileName = arg.copy();
	}
    }

    if ( autoInclude ) {
	int ppos = fileName.findRev('.');
	if ( ppos != -1 && tolower( fileName[ppos + 1] ) == 'h' )
	    noInclude = FALSE;
	else
	    noInclude = TRUE;
    }
    if ( !fileName.isEmpty() && !outputFile.isEmpty() &&
	 includeFile.isEmpty() && includePath.isEmpty() ) {
	includeFile = combinePath(fileName,outputFile);
    }
    if ( includeFile.isEmpty() )
	includeFile = fileName.copy();
    if ( !includePath.isEmpty() ) {
	if ( includePath.right(1) != "/" )
	    includePath += '/';
	includeFile = includePath + includeFile;
    }
    if ( fileName.isNull() && !error ) {
	fileName = "standard input";
	yyin	 = stdin;
    } else if ( argc < 2 || error ) {		// incomplete/wrong args
	fprintf( stderr, "Qt meta object compiler\n" );
	if ( error )
	    fprintf( stderr, "moc: %s\n", error );
	fprintf( stderr, "Usage:  moc [options] <header-file>\n"
		 "\t-o file  Write output to file rather than stdout\n"
		 "\t-i       Do not generate an #include statement\n"
		 "\t-f[file] Force #include, optional file name\n"
		 "\t-p path  Path prefix for included file\n"
		 "\t-k       Do not stop on errors\n"
		 "\t-nw      Do not display warnings\n"
#if 0
		 "\t-ldbg    lex debug output\n"
		 "\t-ydbg    yacc debug output\n"
		 "\t-dbg     test parser, all non-signal members are slots\n"
#endif
		 );
	return 1;
    } else {
	yyin = fopen( (const char *)fileName, "r" );
	if ( !yyin ) {
	    fprintf( stderr, "moc: %s: No such file\n", (const char*)fileName);
	    return 1;
	}
    }
    if ( !outputFile.isEmpty() ) {		// output file specified
	out = fopen( (const char *)outputFile, "w" );	// create output file
	if ( !out ) {
	    fprintf( stderr, "moc: Cannot create %s\n",
		     (const char*)outputFile );
	    return 1;
	}
    } else {					// use stdout
	out = stdout;
    }
    init();
    yyparse();
    fclose( yyin );

    if ( main_gened )
	fprintf(out,
	  "\treturn 0;\n"
	  "}\n");
    if ( !outputFile.isNull() )
	fclose( out );

    if ( !generatedCode && displayWarnings && !mocError ) {
        fprintf( stderr, "%s:%d: Warning: %s\n", fileName.data(), 0,
		 "No relevant classes found. No output generated." );
    }

    slots.clear();
    signals.clear();

    return mocError ? 1 : 0;
}

void replace( char *s, char c1, char c2 )
{
    if ( !s )
	return;
    while ( *s ) {
	if ( *s == c1 )
	    *s = c2;
	s++;
    }
}

/*
  This function looks at two file names and returns the name of the
  infile, with a path relative to outfile. Examples:
    /tmp/abc	/tmp/bcd	->	abc
    xyz/a/bc	xyz/b/ac	->	../a/bc
    /tmp/abc	xyz/klm		-)	/tmp/abc
 */

QCString combinePath( const char *infile, const char *outfile )
{
    QCString a = infile;  replace(a.data(),'\\','/');
    QCString b = outfile; replace(b.data(),'\\','/');
    a = a.stripWhiteSpace();
    b = b.stripWhiteSpace();
    QCString r;
    int i = 0;
    int ncommondirs = 0;
    while ( a[i] && a[i] == b[i] ) {
	if ( a[i] == '/' && i > 0 )
	    ncommondirs++;
	i++;
    }
    if ( ncommondirs > 0 ) {			// common base directory
	while ( i>=0 ) {
	    if ( a[i] == '/' && b[i] == '/' )
		break;
	    --i;
	}
	++i;
	a = &a[i];
	b = &b[i];
    } else {
	if ( (a[0] == '/') || (isalpha(a[0]) && a[1] == ':') )
	    return a;
	b = &b[i];
    }
    i = b.contains('/');
    while ( i-- > 0 )
	r += "../";
    r += a;
    return r;
}


#define getenv hack_getenv			// workaround for byacc
char *getenv()		     { return 0; }
char *getenv( const char * ) { return 0; }


void init()					// initialize
{
    BEGIN OUTSIDE;
    lineNo	 = 1;
    skipClass	 = FALSE;
    skipFunc	 = FALSE;
    staticFunc   = FALSE;
    signals.setAutoDelete( TRUE );
    slots.setAutoDelete( TRUE );

    tmpArgList	 = new ArgList;
    CHECK_PTR( tmpArgList );
    tmpFunc	 = new Function;
    CHECK_PTR( tmpFunc );
}

void initClass()				 // prepare for new class
{
    tmpAccessPerm    = _PRIVATE;
    subClassPerm     = _PRIVATE;
    Q_OBJECTdetected = FALSE;
    skipClass	     = FALSE;
    templateClass    = FALSE;
    slots.clear();
    signals.clear();
    enumDefs.clear();
}


void checkClassDup()
{
  if (skipClass) return;
  for ( QCString *cls = classList.first(); cls ; cls = classList.next() )
    if ( *cls == className )
    {
      skipClass = TRUE;
      return;
    }
  classList.append(new QCString(className));
}

void addSuperClass(char *string)
{
  QCString *super = new QCString(string);
  if ( *super != "Qt" )
    superclassList.append(super);
}

struct NamespaceInfo
{
    QCString name;
    int pLevelOnEntering; // Parenthesis level on entering the namespace
    QDict<char> definedClasses; // Classes defined in the namespace
};

QList<NamespaceInfo> namespaces;

void enterNameSpace( const char *name = 0 )   	 // prepare for new class
{
    static bool first = TRUE;
    if ( first ) {
	namespaces.setAutoDelete( TRUE );
	first = FALSE;
    }

    NamespaceInfo *tmp = new NamespaceInfo;
    if ( name )
	tmp->name = name;
    tmp->pLevelOnEntering = namespacePLevel;
    namespaces.append( tmp );
}

void leaveNameSpace()				 // prepare for new class
{
    NamespaceInfo *tmp = namespaces.last();
    namespacePLevel = tmp->pLevelOnEntering;
    namespaces.remove();
}

QCString nameQualifier()
{
    QListIterator<NamespaceInfo> iter( namespaces );
    NamespaceInfo *tmp;
    QCString qualifier = "";
    for( ; (tmp = iter.current()) ; ++iter ) {
	if ( !tmp->name.isNull() ) {  // If not unnamed namespace
	    qualifier += tmp->name;
	    qualifier += "::";
	}
    }
    return qualifier;
}

int openNameSpaceForMetaObject( FILE *out )
{
    int levels = 0;
    QListIterator<NamespaceInfo> iter( namespaces );
    NamespaceInfo *tmp;
    QCString indent = "";
    for( ; (tmp = iter.current()) ; ++iter ) {
	if ( !tmp->name.isNull() ) {  // If not unnamed namespace
	    fprintf( out, "%snamespace %s {\n", (const char *)indent,
		     (const char *) tmp->name );
	    indent += "    ";
	    levels++;
	}
    }
    QCString nm = className;
    int pos;
    while( (pos = nm.find( "::" )) != -1 ) {
	QCString spaceName = nm.left( pos );
	nm = nm.right( nm.length() - pos - 2 );
	if ( !spaceName.isEmpty() ) {
	    fprintf( out, "%snamespace %s {\n", (const char *)indent,
		     (const char *) spaceName );
	    indent += "    ";
	    levels++;
	}
    }
    return levels;
}

void closeNameSpaceForMetaObject( FILE *out, int levels )
{
    int i;
    for( i = 0 ; i < levels ; i++ )
	    fprintf( out, "}" );
    if ( levels )
	fprintf( out, "\n" );

}

void selectOutsideClassState()
{
    if ( namespaces.count() == 0 )
	BEGIN OUTSIDE;
    else
	BEGIN IN_NAMESPACE;
}

void registerClassInNamespace()
{
    if ( namespaces.count() == 0 )
	return;
    namespaces.last()->definedClasses.insert((const char *)className,(char*)1);
}

//
// Remove white space from SIGNAL and SLOT names.
// This function has been copied from qobject.cpp.
//

inline bool isIdentChar( char x )
{						// Avoid bug in isalnum
    return x == '_' || (x >= '0' && x <= '9') ||
	 (x >= 'a' && x <= 'z') || (x >= 'A' && x <= 'Z');
}

inline bool isSpace( char x )
{
#if defined(_CC_BOR_)
  /*
    Borland C++ 4.5 has a weird isspace() bug.
    isspace() usually works, but not here.
    This implementation is sufficient for our internal use: rmWS()
  */
    return (uchar)x <= 32;
#else
    return isspace( x );
#endif
}

static QCString rmWS( const char *src )
{
    QCString result( strlen(src)+1 );
    char *d = result.data();
    char *s = (char *)src;
    char last = 0;
    while( *s && isSpace(*s) )			// skip leading space
	s++;
    while ( *s ) {
	while ( *s && !isSpace(*s) )
	    last = *d++ = *s++;
	while ( *s && isSpace(*s) )
	    s++;
	if ( *s && isIdentChar(*s) && isIdentChar(last) )
	    last = *d++ = ' ';
    }
    result.truncate( (int)(d - result.data()) );
    return result;
}


void initExpression()
{
    tmpExpression = "";
}

void addExpressionString( char *s )
{
    tmpExpression += s;
}

void addExpressionChar( char c )
{
    tmpExpression += c;
}

void yyerror( char *msg )			// print yacc error message
{
    mocError = TRUE;
    fprintf( stderr, "%s:%d: Error: %s\n", fileName.data(), lineNo, msg );
}

void moc_err( char *s )
{
    yyerror( s );
    if ( errorControl )
	exit( -1 );
}

void moc_err( char *s1, char *s2 )
{
    static char tmp[1024];
    sprintf( tmp, s1, s2 );
    yyerror( tmp );
    if ( errorControl )
	exit( -1 );
}

void moc_warn( char *msg )
{
    if ( displayWarnings )
	fprintf( stderr, "%s:%d: Warning: %s\n", fileName.data(), lineNo, msg);
}

void func_warn( char *msg )
{
    moc_warn( msg );
    skipFunc = TRUE;
}

void operatorError()
{
    if ( !binding )
      moc_warn("Operator functions cannot be signals or slots.");
    skipFunc = TRUE;
}

#ifndef yywrap
int yywrap()					// more files?
{
    return 1;					// end of file
}
#endif

char *stradd( const char *s1, const char *s2 )	// adds two strings
{
    char *n = new char[strlen(s1)+strlen(s2)+1];
    CHECK_PTR( n );
    strcpy( n, s1 );
    strcat( n, s2 );
    return n;
}

char *stradd( const char *s1, const char *s2, const char *s3 )// adds 3 strings
{
    char *n = new char[strlen(s1)+strlen(s2)+strlen(s3)+1];
    CHECK_PTR( n );
    strcpy( n, s1 );
    strcat( n, s2 );
    strcat( n, s3 );
    return n;
}

char *stradd( const char *s1, const char *s2,
	      const char *s3, const char *s4 )// adds 4 strings
{
    char *n = new char[strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+1];
    CHECK_PTR( n );
    strcpy( n, s1 );
    strcat( n, s2 );
    strcat( n, s3 );
    strcat( n, s4 );
    return n;
}


char *straddSpc( const char *s1, const char *s2 )
{
    char *n = new char[strlen(s1)+strlen(s2)+2];
    CHECK_PTR( n );
    strcpy( n, s1 );
    strcat( n, " " );
    strcat( n, s2 );
    return n;
}

char *straddSpc( const char *s1, const char *s2, const char *s3 )
{
    char *n = new char[strlen(s1)+strlen(s2)+strlen(s3)+3];
    CHECK_PTR( n );
    strcpy( n, s1 );
    strcat( n, " " );
    strcat( n, s2 );
    strcat( n, " " );
    strcat( n, s3 );
    return n;
}

char *straddSpc( const char *s1, const char *s2,
	      const char *s3, const char *s4 )
{
    char *n = new char[strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+4];
    CHECK_PTR( n );
    strcpy( n, s1 );
    strcat( n, " " );
    strcat( n, s2 );
    strcat( n, " " );
    strcat( n, s3 );
    strcat( n, " " );
    strcat( n, s4 );
    return n;
}

// Generate C++ code for building member function table

QCString pureClassName()
{
    QCString result;
    int pos = className.findRev( "::");
    if ( pos != -1 )
        result = className.right( className.length() - pos - 2 );
    else
	result = className;
    return result;
}

QCString qualifiedClassName()
{
    QCString tmp = nameQualifier();
    tmp += className;
    return tmp;
}

QCString qualifiedSuperclassName()
{
    if ( namespaces.count() == 0 )
	return superclassName;
    if ( namespaces.last()->definedClasses.find((const char *)superclassName)){
	QCString tmp = nameQualifier();
	tmp += superclassName;
	return tmp;
    } else {
	return superclassName;
    }
}

const int Slot_Num   = 1;
const int Signal_Num = 2;

void generateFuncs( FuncList *list, char *functype, int num )
{
    Function *f;
    for ( f=list->first(); f; f=list->next() ) {
	QCString typstr = "";
	int count = 0;
	Argument *a = f->args->first();
	while ( a ) {
	    if ( !a->leftType.isEmpty() || ! a->rightType.isEmpty() ) {
		if ( count++ )
		    typstr += ",";
		typstr += a->leftType;
		typstr += a->rightType;
	    }
	    a = f->args->next();
	}
	fprintf( out, "    typedef %s(%s::*m%d_t%d)(%s)%s;\n",
		 (const char*)f->type, (const char*)qualifiedClassName(),
		 num, list->at(),
		 (const char*)typstr,  (const char*)f->qualifier );
	f->type = f->name.copy();
	f->type += "(";
	f->type += typstr;
	f->type += ")";
    }
    for ( f=list->first(); f; f=list->next() )
	fprintf( out, "    m%d_t%d v%d_%d = Q_AMPERSAND %s::%s;\n",
		 num, list->at(), num, list->at(),
		 (const char*)qualifiedClassName(), (const char*)f->name);
    if ( list->count() )
	fprintf(out,"    QMetaData *%s_tbl = QMetaObject::new_metadata(%d);\n",
		functype, list->count() );
    for ( f=list->first(); f; f=list->next() )
	fprintf( out, "    %s_tbl[%d].name = \"%s\";\n",
		 functype, list->at(), (const char*)f->type );
    for ( f=list->first(); f; f=list->next() )
	fprintf( out, "    %s_tbl[%d].ptr = *((QMember*)&v%d_%d);\n",
		 functype, list->at(), num, list->at() );
}

bool checkClassWarnings()
{
  if ( !Q_OBJECTdetected ) {
    // Public methods are stored in singles
    if ( slots.count() == 0 )
      return false;
    generatedCode = TRUE;
    if ( !binding && displayWarnings )
      moc_err("The declaration of the class \"%s\" contains slots "
	      "and/or signals\n\t but no Q_OBJECT macro!", className.data());
  } else {
    if ( superclassName.isEmpty() )
      moc_err("The declaration of the class \"%s\" contains the\n"
	      "\tQ_OBJECT macro but does not inherit from any class!\n"
	      "\tInherit from QObject or one of its descendants"
	      " or remove Q_OBJECT. ", className.data() );
  }
  if ( templateClass ) {			// don't generate for class
    moc_err( "Sorry, Qt does not support templates that contain\n"
	     "signals, slots or Q_OBJECT. This will be supported soon." );
    return false;
  }

  generatedCode = TRUE;
  return true;
}

class QTypeMap : public QAsciiDict<QString> {
 public:
  void loadFrom(const char *fileName);
};

void QTypeMap::loadFrom(const char *filename)
{
  char buffer[1024];
  FILE *infile;

  infile = fopen(filename, "r");
  if ( !infile ) {
    fprintf(stderr, "Warning: file %s not found.\n", filename);
    return;
  }
  while ( !feof(infile) ) {
    fgets(buffer, sizeof(buffer), infile);
    QString line(buffer);

    static QString str("\"\[^\"]*\"");
    static QString eq("[ \t]*=[ \t]*");
    static QRegExp r_str(str);
    static QRegExp r_eq(str);

    int f_pos, f_len;
    int e_pos;
    int s_pos, s_len;
    if ( !line.isEmpty() ) {
      // find the first entry
      if ( (f_pos = r_str.match(line, 0, &f_len, FALSE)) < 0 )
	goto skip;
      if ( (e_pos = r_eq.match(line, f_pos + f_len, NULL, FALSE)) < 0 )
	goto skip;
      if ( (s_pos = r_eq.match(line, e_pos, &s_len, FALSE)) < 0 )
	goto skip;

      QString first = line.mid(f_pos + 1, f_len - 2); // ignore "s
      QString second = line.mid(s_pos + 1, s_len - 2);

      insert(first, new QString(second));
    }
skip: ;
  }
  fclose(infile);
}

static QTypeMap ctypeMap;
static bool ctypeMapLoaded = FALSE;

void loadCMappingFile()
{
  if ( ctypeMapLoaded ) return;
  ctypeMap.loadFrom("cify_types");
  ctypeMapLoaded = TRUE;
}

static QTypeMap ccastMap;
static bool ccastMapLoaded = FALSE;

void loadCCastMappingFile()
{
  if ( ccastMapLoaded ) return;
  ccastMap.loadFrom("cast_types");
  ccastMapLoaded = TRUE;
}

static QTypeMap pascalTypeMap;
static bool pascalMapLoaded = FALSE;

void loadPascalMappingFile()
{
  if ( pascalMapLoaded ) return;
  pascalTypeMap.loadFrom("pascalify_types");
  pascalMapLoaded = TRUE;
}

class SpecialCopy {
  public:
    SpecialCopy(QCString qtType, QCString bindType,
      QCString bindToQt, QCString qtToBind, QCString init, QCString final,
      bool in, bool out, bool address)
    : qtType(qtType), bindType(bindType), bindToQt(bindToQt),
      qtToBind(qtToBind), in(in), out(out), address(address),
      init(init), final(final)
    {
      needTemp = !bindToQt.isEmpty() || !qtToBind.isEmpty();
    }

    QCString qtType;
    QCString bindType;
    QCString bindToQt;
    QCString qtToBind;
    QCString init;
    QCString final;
    bool     in;
    bool     out;
    bool     address;
    bool     needTemp;
};

class SpecialCopiesMap : public QAsciiDict<SpecialCopy> {
 public:
  void loadFrom(const char *fileName);
};

bool find_ident(const QString& line, const QRegExp& exp, int &p, int &pos,
  int &len)
{
  bool result = (pos = exp.match(line, p, &len, FALSE)) < 0;
  p = pos + len;

  // ignore the "'s
  pos++;
  len -= 2;
  return result;
}

bool skip(const QString& line, const QRegExp& exp, int &pos)
{
  int p, len;
  bool result = (p = exp.match(line, pos, &len, FALSE)) < 0;
  pos = p + len;
  return result;
}

bool ignore_line(const char *buffer)
{
  if ( buffer[0] == '#' || buffer[0] == '\0' ) return TRUE;
  for ( const char *p = buffer; *p; p++ )
    if ( *p != ' ' && *p != '\n' && *p != '\t' ) return FALSE;
  return TRUE;
}

QCString& cifyType(QCString& type);

void SpecialCopiesMap::loadFrom(const char *filename)
{
  int line_no;
  char buffer[1024];
  FILE *infile;

  infile = fopen(filename, "r");
  if ( !infile ) {
    fprintf(stderr, "Warning: file %s not found.\n", filename);
    return;
  }
  line_no = 0;
  while ( !feof(infile) ) {
    fgets(buffer, sizeof(buffer), infile);
    line_no++;
    if ( ignore_line(buffer) ) continue;
    QString line(buffer);

    static QString str("\"\[^\"]*\"");
    static QRegExp r_str(str);
    static QString maps("[ \t]*=>[ \t]*");
    static QRegExp r_maps(maps);
    static QString comma("[ \t]*,[ \t]*");
    static QRegExp r_comma(comma);

    int qtType_pos, qtType_len;
    int bindType_pos, bindType_len;
    int from_pos, from_len;
    int to_pos, to_len;
    // optionals need to be initialized
    int init_pos = 0;
    int init_len = 0;
    int final_pos = 0;
    int final_len = 0;

    int p = 0;

    // find qtType
    if ( find_ident(line, r_str, p, qtType_pos, qtType_len) )
      goto error;
    if ( skip(line, r_maps, p) )
      goto error;

    // find bindType
    if ( find_ident(line, r_str, p, bindType_pos, bindType_len) )
      goto error;
    if ( skip(line, r_comma, p ) )
      goto error;

    // find from
    if ( find_ident(line, r_str, p, from_pos, from_len) )
      goto error;
    if ( skip(line, r_comma, p) )
      goto error;

    // find to
    if ( find_ident(line, r_str, p, to_pos, to_len) )
      goto error;

    // The rest of the fields are optional

    // find init
    if ( skip(line, r_comma, p) )
      goto done;
    if ( find_ident(line, r_str, p, init_pos, init_len) )
      goto error;

    // find final
    if ( skip(line, r_comma, p) )
      goto done;
    if ( find_ident(line, r_str, p, final_pos, final_len) )
      goto error;

    // although techincally not part of the syntax, this ignores garbage at
    // the end of a line. A file should not depend on this.

    done:
    {
      QCString qtType = (const char *)line.mid(qtType_pos, qtType_len);
      QCString bindType = (const char *)line.mid(bindType_pos, bindType_len);
      QCString from = (const char *)line.mid(from_pos, from_len);
      QCString to = (const char *)line.mid(to_pos, to_len);
      QCString init, final;
      if ( init_pos )
	init = (const char *)line.mid(init_pos, init_len);
      if ( final_pos )
	final = (const char *)line.mid(final_pos, final_len);

      insert(qtType, new SpecialCopy(qtType, bindType, to, from, init, final, TRUE, FALSE, FALSE));
      QCString nType;
      nType = "const " + qtType; insert(nType, new SpecialCopy(nType, bindType, to, from, init, final, TRUE, FALSE, FALSE));

      if ( to_len || from_len ) {
	nType = qtType + "*"; insert(nType, new SpecialCopy(nType, bindType, to, from, init, final, TRUE, TRUE, TRUE));
	nType = qtType + "&"; insert(nType, new SpecialCopy(nType, bindType, to, from, init, final, TRUE, TRUE, FALSE));
	nType = "const " + qtType + "*"; insert(nType, new SpecialCopy(nType, bindType, to, from, init, final, TRUE, FALSE, TRUE));
	nType = "const " + qtType + "&"; insert(nType, new SpecialCopy(nType, bindType, to, from, init, final, TRUE, FALSE, FALSE));
      } else {
	QCString s = qtType + "*";
	cifyType(s);
	if ( s != bindType )
	  pascalTypeMap.insert(s, new QString(bindType));
      }
    }
    continue;

error:
    fprintf(stderr, "Warning: incorrect syntax in special_copys on line %d.\n",
      line_no);
  }
  fclose(infile);
}

static SpecialCopiesMap specialCopies;
static bool specialCopiesLoaded = FALSE;

void loadSpecialCopiesFile()
{
  if ( specialCopiesLoaded ) return;
  specialCopies.loadFrom("special_copies");
  specialCopiesLoaded = TRUE;
}

QCString& replaceAmpersands(QCString& type)
{
  int pos;

  // turn all '&'s into '*'s
  while ( (pos = type.find('&')) >= 0 ) {
    type.replace(pos, 1, "*");
  }

  return type;
}

QCString& cifyType(QCString& type)
{
  int pos, len;

  replaceAmpersands(type);

  QString stype(type);

  // remove all prefixes
  static QString type_pre("[A-Za-z0-9]*[ \t]*::");
  static QRegExp reg_type_pre(type_pre);

  while ( (pos = reg_type_pre.match(stype, 0, &len, false) ) >= 0 )
    stype.remove(pos, len);

  // turn all Qxxx * to QxxxH types
  static QString qtype_ptr("Q[A-Za-z0-9]*[ \t]*\\*");
  static QString qtype_only("Q[A-Za-z0-9]*");
  static QRegExp reg_qtype_ptr(qtype_ptr);
  static QRegExp reg_qtype_only(qtype_only);

  if ( ( pos = reg_qtype_ptr.match(stype, 0, &len, false) ) >= 0 ) {
    int q_type_len;
    reg_qtype_only.match(stype, pos, &q_type_len, true);
    stype.replace(pos + q_type_len, len - q_type_len, "H ");
  }

  // test for simple replacements
  type = stype.stripWhiteSpace();
  loadCMappingFile();
  if ( QString *mappedType = ctypeMap.find((const char *)type) )
    type = (const char *)*mappedType;

  return type;
}

QCString& pascalifyType(QCString& type)
{
  cifyType(type);

  int pos, len;
  QString stype(type);

  // remove const or volatile's
  static QString s_const("[\t ]*const[\t ]*");
  static QRegExp r_const(s_const);
  static QString s_volatile("[\t ]*volatile[\t ]*");
  static QRegExp r_volatile(s_volatile);
  while ( (pos = r_const.match(stype, 0, &len, false) ) >= 0 )
    stype.remove(pos, len);
  while ( (pos = r_volatile.match(stype, 0, &len, false) ) >= 0 )
    stype.remove(pos, len);

  // remove spaces before *'s
  static QString s_star("[\t ]+\\*");
  static QRegExp r_star(s_star);
  while( (pos = r_star.match(stype, 0, &len, false) ) >= 0 )
    stype.remove(pos, len - 1);

  // map the type
  loadPascalMappingFile();
  if ( QString *mappedType = pascalTypeMap.find((const char *)stype) )
    stype = (const char *)*mappedType;

  type = stype;

  return type;
}

QCString cCast(QCString type)
{
  QCString cast;

  type = type.stripWhiteSpace();
  loadCCastMappingFile();
  if ( QString *mappedCast = ccastMap.find((const char *)type) ) {
    if ( !mappedCast->isEmpty() )
      cast = (const char *)mappedCast;
    else
      cast = "";
  }  else {
    cast = "";
    if ( type.find('&') >= 0 ) {
      replaceAmpersands(type);
      cast += "*";
    }
    cast += "(" + type + ")";
  }

  return cast;
}

QCString baseTypeOf(QCString type)
{
  static QString s_const("[\t ]*const[\t ]*");
  static QRegExp r_const(s_const);
  static QString s_ident("[A-Za-z_][A-Za-z0-9_<>]*");
  static QRegExp r_ident(s_ident);

  QString s(type);
  int pos, len;
  while ( (pos = r_const.match(s, 0, &len, false) ) >= 0 )
    s.remove(pos, len);
  pos = r_ident.match(s, 0, &len, false);
  if ( pos < 0 ) {
    fprintf(stderr, "Warning: no base type found in type %s.\n",
      (const char *)type);
    s = "";
  }
  else
    s = s.mid(pos, len);
  QCString ret(s);
  return ret;
}

struct Overload {
  int count;
  int total;

  Overload() : count(1), total(1) {}
};

class OverloadMap : public QDict<Overload> {
 public:
  OverloadMap() : QDict<Overload>() {}
  int insertMethod(const QCString &method_name);
  bool overloaded(const QCString &method_name);
  void zero();
};

int OverloadMap::insertMethod(const QCString &method_name)
{
  Overload *overload;

  QString method_name_s((const char *)method_name);
  if ( (overload = (*this)[method_name_s]) != NULL ) {
    if ( overload->count == overload->total )
      overload->total++;
    return ++overload->count;
  } else {
    insert(method_name_s, new Overload());
    return 0;
  }
}

bool OverloadMap::overloaded(const QCString &method_name)
{
  QString method_name_s((const char *)method_name);
  Overload *overload;
  if ( (overload = (*this)[method_name_s]) != NULL )
    return overload->total > 1;
  return FALSE;
}

void OverloadMap::zero()
{
  Overload *o;

  for ( QDictIterator<Overload> it(*this); it.current(); ++it )
    it.current()->count = 0;
}

void generateHeader(const char *hdr1, const char *hdr2, const char *hdr3,
		    const char *hdr4)
{
  QDateTime dt = QDateTime::currentDateTime();
  QCString dstr = dt.toString().ascii();
  QCString fn = fileName;
  int i;
  static bool generated;

  if ( binding && generated ) return;
  i = fileName.length()-1;
  while ( i>0 && fileName[i-1] != '/' && fileName[i-1] != '\\' )
    i--;				// skip path
  if ( i >= 0 )
    fn = &fileName[i];
  fprintf( out, hdr1, (const char*)qualifiedClassName(),(const char*)fn);
  fprintf( out, hdr2, (const char*)dstr );
  fprintf( out, hdr3 );
  fprintf( out, hdr4 );
  generated = TRUE;
}

/*
enum GenOptions { dont_gen_types = 0, get_vars = 0, gen_types = 0x1,
  gen_trail_comma = 0x2, cify_types = 0x4, gen_casts = 0x8,
  dont_gen_vars = 0x10, gen_lead_comma = 0x20 };

void generateCParams(FILE *out, Function *f, GenOptions options)
{
  bool first = TRUE;
  bool gen_lead = options & gen_lead_comma;
  int count = 0;
  QCString name("");
  QCString cast("");
  const char *sep = (options & dont_gen_vars) ? "" : " ";

  for ( Argument *a = f->args->first(); a; a = f->args->next() ) {
    if ( !first || gen_lead ) fprintf(out, ", ");
    first = FALSE;
    if ( a->name.isEmpty() ) {
      char buf[10];
      sprintf(buf, "p%d", count);
      a->name = buf;
    }
    QCString cpptype(a->leftType + a->rightType);
    QCString type(cpptype);
    if ( options & gen_casts )
      cast = cCast(cpptype);
    if ( options & gen_types ) {
      if ( options & cify_types )
        cifyType(type);
      if ( !(options & dont_gen_vars ) )
        name = a->name;
      fprintf(out, "%s%s%s%s",
        (const char *)type,
        sep,
        (const char *)cast,
	(const char *)name);
    } else
      fprintf(out, "%s%s",
	(const char *)cast,
	(const char *)a->name);
  }
  if ( !first && (options & gen_trail_comma) )
    fprintf(out, ", ");
}
*/

class BindArgument {
  public:
    BindArgument(Argument *a) : a(a), copy(NULL) {
      if ( a )
	name = a->name;
    }

    Argument    *a;
    QCString    name;
    QCString	expr;
    QCString    qtType;
    QCString    cType;
    QCString    pascalType;
    SpecialCopy *copy;

    bool is_ret() { return a == NULL; } // a NULL a indicates out retval arg
};

class BindArgList : public QList<BindArgument>
{
  public:
    BindArgList() { setAutoDelete(TRUE); }
};

class BindFunction
{
  public:
    BindFunction(Function *f) : func(f), ret_arg(NULL),
      is_func(TRUE), flatNumber(0), is_static(f && f->is_static),
      is_overloaded(FALSE) {
      if ( f )
	name = f->name;
    }

    QCString              name;
    Function              *func;
    BindArgList           args;
    QCString              flatName;
    int                   flatNumber;
    BindArgument         *ret_arg;
    bool                  is_func;
    bool                  is_static;
    bool		  is_overloaded;
    QCString              ret_qtType;
    QCString              ret_cType;
    QCString              ret_pascalType;
};

class BindFuncList : public QList<BindFunction>
{
  public:
    BindFuncList() { setAutoDelete(TRUE); }
};

BindFuncList bindings;

/* Builds the binding function list. In building this list it,
 *  - Calculates what the flattened name should be (i.e. the non-overloaded
 *    name).
 *  - Determines if the return value or parameters need a special copy.
 *    - The return parameter is turned into an argument if it needs to be
 *      copied.
 *  - Assigns a unique argument name if one is not present.
 */
void buildBindings(FuncList &funcs)
{
  OverloadMap map;
  Function *f;

  // pre-build overload map to easily detect overloaded methods
  for ( f=funcs.first(); f; f=funcs.next() )
    if ( f->accessPerm == _PUBLIC )
      map.insertMethod(f->name);
  map.zero();

  bindings.clear();
  for ( f=funcs.first(); f; f=funcs.next() ) {
    if ( f->accessPerm != _PUBLIC ) continue;

    // Associate binding information with the function itself
    BindFunction *bf = new BindFunction(f);
    bindings.append(bf);

    // Calculate the flat name (i.e. the non-overloaded name) of the function
    QCString method_name(f->name);
    int method_num;
    if ( (method_num = map.insertMethod(method_name)) > 0 ) {
      if ( method_num == 1 )
	method_num = 0;
      else {
	QCString method_num_s;
	method_num_s.setNum(method_num);
	method_name = method_name + method_num_s;
      }
    }
    bf->flatName = method_name;
    bf->flatNumber = method_num;
    bf->is_overloaded = map.overloaded(f->name);

    // Calculate return value information
    loadSpecialCopiesFile();
    SpecialCopy *ret_copy = f->type != f->name ? specialCopies.find(f->type) : NULL;
    if ( ret_copy ) {
      // Return value requires a special copy, make it an out parameter instead
      //  turn function into a procedure
      bf->is_func = FALSE;
      bf->ret_cType = "void";

      //  return result as an out parameter
      BindArgument *ret_val = new BindArgument(NULL);
      ret_val->name = "retval";
      bf->args.append(ret_val);
      ret_val->copy = ret_copy;
      ret_val->qtType = f->type;
      ret_val->cType = ret_copy->bindType;
      ret_val->pascalType = ret_copy->bindType;
      bf->ret_arg = ret_val;
    } else if ( f->type == "void" ) {
      // really a procedure
      bf->is_func = FALSE;
      bf->ret_cType = "void";
    } else {
      // collect return information
      bf->ret_qtType = f->type;
      bf->ret_cType = f->type;
      bf->ret_pascalType = f->type;
      cifyType(bf->ret_cType);
      pascalifyType(bf->ret_pascalType);
    }

    // Associate binding information with the arguments
    int count = 1;
    QCString type;
    for ( Argument *a = f->args->first(); a; a = f->args->next(), count++ ) {
      if ( !a->leftType.isEmpty() || ! a->rightType.isEmpty() ) {
	BindArgument *ba = new BindArgument(a);
	bf->args.append(ba);

	// determine the qtType
	type = a->leftType;
	type += a->rightType;
	ba->qtType = type;

	// look for special copies
	SpecialCopy *s = specialCopies.find(type);
	ba->copy = s;

	// determine the binding types
	ba->cType = s ? s->bindType : cifyType(type);
	ba->pascalType = s ? s->bindType : pascalifyType(type);

	// give anonomous parameters a name
	if ( ba->name.isEmpty() ) {
	  char buf[20];
	  sprintf(buf, "p%d", count);
	  ba->name = buf;
	}
      }
    }
  }
}

enum GenOptions {
  gen_types         = 0x001,
  gen_vars          = 0x002,
  gen_trail_comma   = 0x004,
  gen_lead_comma    = 0x008,
  cify_types        = 0x010,
  gen_casts         = 0x020,
  gen_ret_val       = 0x040,
  gen_exprs    	    = 0x080,
  gen_bind_type	    = 0x100
};

void generateCParams(FILE *out, BindFunction *f, GenOptions options)
{
  bool first = FALSE;
  bool gen_comma = options & gen_lead_comma;
  QCString name("");
  QCString cast("");
  QCString type("");
  const char *sep = (options & gen_vars) && (options & gen_types) ? " " : "";

  for ( BindArgument *a = f->args.first(); a; a = f->args.next() ) {
    if ( !(options & gen_ret_val) && a->is_ret() )
      continue;
    if ( gen_comma ) fprintf(out, ", ");
    gen_comma = TRUE;
    first = FALSE;

    if ( options & gen_casts && !a->copy && a->expr.isEmpty() && a->qtType != a->cType)
      cast = cCast(options & gen_bind_type ? a->cType : a->qtType);
    else
      cast = "";
    if ( options & gen_vars )
      if ( options & gen_exprs && !a->expr.isEmpty()  )
	name = a->expr;
      else
	name = a->name;
    if ( options & gen_types )
      type = options & cify_types ? a->cType : a->qtType;
    fprintf(out, "%s%s%s%s",
      (const char *)type,
      sep,
      (const char *)cast,
      (const char *)name);
  }
  if ( !first && (options & gen_trail_comma) )
    fprintf(out, ", ");
}

#define EXPORT_NAME "C_EXPORT"

typedef void (*CallPrefix)(BindFunction *f, GenOptions &options);

void genConstructorCallPrefix(BindFunction *, GenOptions &)
{
  fprintf(out,	"(%sH) new %s(",
	(const char *)className,
	(const char *)className);
}

void genCallPrefix(BindFunction *f, GenOptions &)
{
  QCString cast("");

  if ( f->is_func ) {
    cast = cCast(f->ret_cType) + " ";
    if ( !f->ret_arg && f->ret_qtType.find('&') >= 0 )
      cast = cast + "&";
  }
  if ( f->is_static )
    fprintf(out, "%s%s::%s(",
	(const char *)cast,
	(const char *)className,
	(const char *)f->name);
  else
    fprintf(out, "%s((%s *)handle)->%s(",
	  (const char *)cast,
	  (const char *)className,
	  (const char *)f->name);
}

void genCHookPrefix(BindFunction *f, GenOptions &options)
{
  fprintf(out, "(*(func_type)%s_event.func)(%s_event.data",
    (const char *)f->flatName, (const char *)f->flatName);
  options = GenOptions(options | gen_lead_comma);
}

enum CallDir { cdBindToQt, cdQtToBind };

void genCCall(BindFunction *f, CallDir dir, CallPrefix prefix)
{
  // Generate temporaries for copies
  bool ins = FALSE;
  bool outs = FALSE;
  bool inits = FALSE;
  for (BindArgument *a = f->args.first(); a; a = f->args.next() ) {
    if ( a->copy ) {
      if ( a->copy->needTemp ) {
	fprintf(out, "\t%s t_%s;\n",
	  dir == cdBindToQt ? (const char *)baseTypeOf(a->qtType) :
	    (const char *)a->cType, (const char *)a->name);
	ins = ins || (a->copy->in && a != f->ret_arg);
	outs = outs || (a->copy->out || a == f->ret_arg);
	if ( dir == cdQtToBind )
	  inits = inits || !a->copy->init.isEmpty();
	if ( a->copy->address && dir == cdBindToQt )
	  a->expr = a->name + " ? &t_" + a->name + " : NULL";
	else
	  a->expr = "t_" + a->name;
      } else
	if ( dir == cdBindToQt )
	  a->expr = (a->copy->address ? "" : "*") + ("(" + baseTypeOf(a->qtType) + " *)" + a->name);
	else
	  a->expr = cCast(a->cType) + (a->copy->address ? "" : "&") + a->name;
    }
    else if ( dir == cdQtToBind && a->qtType.find('&') >= 0 )
      a->expr = cCast(a->cType) + "&" + a->name;
  }

  if ( inits )
    for (BindArgument *a = f->args.first(); a; a = f->args.next() )
      if ( a->copy && a->copy->needTemp && !a->copy->init.isEmpty() )
	fprintf(out, "\t%s(t_%s);\n", (const char *)a->copy->init, (const char *)a->name);

  // Generate the in copies
  if ( ins )
    for (BindArgument *a = f->args.first(); a; a = f->args.next() ) {
      if ( a->copy && a->copy->needTemp && a->copy->in && a != f->ret_arg ) {
	if ( dir == cdBindToQt ) {
	  if ( a->copy->address )
	    fprintf(out, "\tif ( %s )\n\t",
	      (const char *)a->name);
	  if ( a->copy->bindToQt.isEmpty() )
	    fprintf(out, "\tt_%s = *(%s *)%s;\n",
	      (const char *)a->name,
	      (const char *)baseTypeOf(a->qtType),
	      (const char *)a->name);
	  else
	    fprintf(out, "\t%s(%s, t_%s);\n",
	      (const char *)a->copy->bindToQt,
	      (const char *)a->name,
	      (const char *)a->name);
	} else {
	  if ( a->copy->bindToQt.isEmpty() )
	   fprintf(out, "\tt_%s = (%s)%s%s;",
	     (const char *)a->name,
	     (const char *)a->cType,
	     a->copy->address ? "" : "&",
	     (const char *)a->name);
	  else
	    fprintf(out, "\t%s(%s, t_%s);\n",
	      (const char *)a->copy->qtToBind,
	      (const char *)a->name,
	      (const char *)a->name);
	}
      }
    }

  // Generate the call
  if ( f->ret_arg )
    if ( f->ret_arg->copy->needTemp )
      fprintf(out, "\tt_%s = ", (const char *)f->ret_arg->name);
    else
      fprintf(out, "\t*(%s *)%s = ",
	(const char *)baseTypeOf(f->ret_arg->qtType),
	(const char *)f->ret_arg->name);
  else if ( outs && f->is_func )
    fprintf(out, "\t%s t_retval;\n"
		 "\tt_retval = ",
      (const char *)f->ret_cType);
  else if ( f->is_func || f->name == className )
    fprintf(out, "\treturn ");
  else
    fprintf(out, "\t");

  GenOptions options = GenOptions(gen_vars | gen_casts | gen_exprs);
  if ( dir == cdQtToBind )
    options = GenOptions(options | gen_bind_type);
  prefix(f, options);
  generateCParams(out, f, options);
  fprintf(out, ");\n");

  // Generate the out copies
  if ( outs ) {
    for (BindArgument *a = f->args.first(); a; a = f->args.next() )
      if ( a->copy && a->copy->needTemp && (a->copy->out || a == f->ret_arg) ) {
	if ( dir == cdBindToQt ) {
	  if ( a->copy->address )
	    fprintf(out, "\tif ( %s )\n\t",
	      (const char *)a->name);
	  if ( !a->copy->qtToBind.isEmpty() )
	    fprintf(out, "\t%s(t_%s, %s);\n",
	      (const char *)a->copy->qtToBind,
	      (const char *)a->name,
	      (const char *)a->name);
	} else {
	  if ( !a->copy->qtToBind.isEmpty() )
	    fprintf(out, "\t%s(t_%s, %s%s);\n",
	      (const char *)a->copy->bindToQt,
	      (const char *)a->name,
	      a->copy->address ? "*" : "",
	      (const char *)a->name);
	}
      }
  }

  if ( inits )
    for (BindArgument *a = f->args.first(); a; a = f->args.next() )
      if ( a->copy && a->copy->needTemp && !a->copy->final.isEmpty() )
	fprintf(out, "\t%s(t_%s);\n", (const char *)a->copy->final, (const char *)a->name);

  // if ret was not turned into a parameter, return it now.
  if ( outs && !f->ret_arg && f->is_func )
    fprintf(out, "\treturn t_retval;\n");
}

void generateCBinding()
{
  char *hdr1 =
    "/****************************************************************************\n"
    "** %s C bindings generated from reading C++ file '%s'\n**\n";
  char *hdr2 =
    "** Created: %s\n"
    "**      by: The Qt Meta Object Compiler ($Revision: 1.1 $)\n**\n";
  char *hdr3 =
    "** WARNING! All changes made in this file will be lost!\n";
  char *hdr4 =
    "*****************************************************************************/\n\n";


  checkClassDup();

  if ( skipClass )
    return;

  if ( !checkClassWarnings() )
    return;

  generateHeader(hdr1, hdr2, hdr3, hdr4);

  bool dtor_emitted = FALSE;
  const char *class_name = (const char *)className;

  if ( cTypes ) {
    // declare the class' handle type
    fprintf(out, "typedef struct %s__ { int dummy; } *%sH;\n\n", class_name,
	    class_name);
    // declare all supported enumerations
    bool enum_gened = FALSE;
    for ( EnumDef *e = enumDefs.first(); e; e = enumDefs.next() ) {
      bool first = true;
      if ( e->assigned )
	fprintf(out, "/* !!! The values of %s are not correct !!! */\n",
	  (const char *)e->name);
      fprintf(out, "typedef enum { ", (const char *)e->name);
      for ( EnumValue *ev = e->first(); ev; ev = e->next() ) {
	if ( !first )
	  fprintf(out, ", ");
	first = false;
	fprintf(out, "%s", (const char *)ev->name);
      }
      fprintf(out, " } %s;", (const char *)e->name);
      fprintf(out, "\n");
      enum_gened = TRUE;
    }
    if ( enum_gened )
      fprintf(out, "\n");
  }

  if ( cHeader ) {
    buildBindings(slots);
    BindFunction *f;
    for ( f = bindings.first(); f; f = bindings.next() ) {
      if ( f->name == className ) {
	// constructor, emit constructor and destructor
	if ( !f->flatNumber )
	  fprintf(out, EXPORT_NAME " %sH %s_create(", class_name,
	    class_name);
	else
	  fprintf(out, EXPORT_NAME " %sH %s_create%d(", class_name,
	    class_name, f->flatNumber);
	generateCParams(out, f, GenOptions(gen_types | gen_vars | cify_types));
	fprintf(out, ");\n");

	if ( !dtor_emitted ) {
	  fprintf(out, EXPORT_NAME " void %s_destroy(%sH handle);\n",
	    class_name, class_name);
	  dtor_emitted = TRUE;
	}
      } else {
	// normal method
	 if ( f->is_static )
	  // static method
	  fprintf(out, EXPORT_NAME " %s %s_%s(",
	    (const char *)f->ret_cType,
	    (const char *)class_name,
	    (const char *)f->flatName);
	 else
	   // non-static method
	   fprintf(out, EXPORT_NAME " %s %s_%s(%sH handle",
		  (const char *)f->ret_cType,
		  (const char *)class_name,
		  (const char *)f->flatName,
		  (const char *)class_name);
	 generateCParams(out, f, GenOptions(gen_types | gen_vars | cify_types |
	 gen_ret_val | (f->is_static ? 0 : gen_lead_comma) ));
	 fprintf(out, ");\n");
      }
    }

    if ( superclassList.count() > 1 ) {
      // More than one super-class, gen a conversion functions
      QCString *base = superclassList.first();
      for ( base = superclassList.next(); base; base = superclassList.next() ) {
	fprintf(out, EXPORT_NAME " %sH %s_to_%s(%sH handle);\n",
	  (const char *)*base,
	  class_name,
	  (const char *)*base,
	  class_name);
      }
    }


    if ( cHook ) {
      OverloadMap sig_map;
      for ( Function *f = signals.first(); f; f = signals.next() ) {
	char method_num_s[10];
	int method_num;
	if ( (method_num = sig_map.insertMethod(f->name)) > 1 )
	  sprintf(method_num_s, "%d", method_num);
	else
	  method_num_s[0] = '\0';
	fprintf(out, EXPORT_NAME " void %s_hook_%s%s(%sH handle, QHook hook);\n",
	  class_name,
	  (const char *)f->name,
	  method_num_s,
	  class_name);
      }
    }
  }

  if ( cImpl ) {
    buildBindings(slots);
    dtor_emitted = FALSE;
    for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
      if ( f->name == className ) {
	// constructor, emit constructor and destructor
	if ( !f->flatNumber )
	  fprintf(out, "%sH %s_create(", class_name,
	    class_name);
	else
	  fprintf(out, "%sH %s_create%d(", class_name,
	    class_name, f->flatNumber);
	generateCParams(out, f, GenOptions(gen_types | gen_vars | cify_types));
	fprintf(out, ")\n");
	fprintf(out, "{\n");
        genCCall(f, cdBindToQt, genConstructorCallPrefix);
	fprintf(out, "}\n\n");

	if ( !dtor_emitted ) {
	  fprintf(out, "void %s_destroy(%sH handle)\n"
		  "{\n"
		  "\tdelete (%s *)handle;\n"
		  "}\n\n",
	    class_name,
	    class_name,
	    class_name);
	  dtor_emitted = TRUE;
	}
      } else {
        // Generate function definition
	if ( f->is_static )
	  fprintf(out, "%s %s_%s(",
		  (const char *)f->ret_cType,
		  class_name,
		  (const char *)f->flatName);
	else
	  fprintf(out, "%s %s_%s(%sH handle",
		  (const char *)f->ret_cType,
		  class_name,
		  (const char *)f->flatName,
		  class_name);
	generateCParams(out, f, GenOptions(gen_types | gen_vars | cify_types |
	  gen_ret_val | (f->is_static ? 0 : gen_lead_comma)));
	fprintf(out, ")\n{\n");

        // Generate call
        genCCall(f, cdBindToQt, genCallPrefix);
	fprintf(out, "}\n\n");
      }
    }

    if ( superclassList.count() > 1 ) {
      // More than one super-class, gen a conversion functions
      QCString *base = superclassList.first();
      for ( base = superclassList.next(); base; base = superclassList.next() ) {
	fprintf(out, "%sH %s_to_%s(%sH handle)\n"
                     "{\n"
                     "\treturn (%sH)(%s *)(%s *)handle;\n"
                     "}\n\n",
	  (const char *)*base,
	  class_name,
	  (const char *)*base,
	  class_name,
	  (const char *)*base,
	  (const char *)*base,
	  class_name);
      }
    }

  }

  if ( cHook && !superclassName.isEmpty() ) {
    static bool header_generated = false;
    OverloadMap sig_map;

    // Include file
    if ( !noInclude & !header_generated ) {
      fprintf( out, "#include \"%s\"\n\n", (const char*)includeFile );
      header_generated = true;
    }

    // Declare hook class
    QString superName;
    if ( superclassList.first() )
      superName = *superclassList.first();
    else
      superName = superclassName;
    fprintf(out, "class %s_hook : public %s_hook {\n"
		 "  Q_OBJECT\n"
		 "  public:\n"
		 "    %s_hook(QObject *handle) : %s_hook(handle) {\n",
      class_name,
      (const char *)superName,
      class_name,
      (const char *)superName);
    // initialize hooks to NULL
    for ( Function *f = signals.first(); f; f = signals.next() ) {
      char method_num_s[10];
      int method_num;
      if ( (method_num = sig_map.insertMethod(f->name)) > 1 )
	sprintf(method_num_s, "%d", method_num);
      else
	method_num_s[0] = '\0';
      fprintf(out, "      %s%s_event.func = NULL;\n",
	(const char *)f->name, method_num_s);
    }
    fprintf(out, "    }\n");


    // create hook functions
    buildBindings(signals);
    bool first = TRUE;
    for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
      // connect hook if the assigned value is non NULL and assign the hook
      fprintf(out, "    void hook_%s(QHook &hook) { \n"
		   "      if ( !%s_event.func )\n"
		   "        connect(handle, SIGNAL(%s(",
	(const char *)f->flatName,
	(const char *)f->flatName,
	(const char *)f->name);
      generateCParams(out, f, GenOptions(gen_types));
      fprintf(out, ")), this, SLOT(%s_hook(", (const char *)f->flatName);
      generateCParams(out, f, GenOptions(gen_types));
      fprintf(out, ")));\n"
		   "      %s_event = hook;\n",
	(const char *)f->flatName);
      // disconnect if the event is assign NULL
      fprintf(out, "      if ( !hook.func )\n"
		   "        disconnect(handle, SIGNAL(%s(",
	(const char *)f->name);
      generateCParams(out, f, GenOptions(gen_types));
      fprintf(out, ")), this, SLOT(%s_hook(",
	(const char *)f->flatName);
      generateCParams(out, f, GenOptions(gen_types));
      fprintf(out, ")));\n"
		   "    }\n");

      first = FALSE;
    }
    if ( !first ) fprintf(out, "\n");

    // Declare all event hooks
    first = TRUE;
    for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
      if ( first )
	fprintf(out, "  private slots:\n");
      fprintf(out, "    void %s_hook(", (const char *)f->flatName);
      generateCParams(out, f, GenOptions(gen_types | gen_vars));
      // finish delcaration and start function pointer decl
      fprintf(out, ") {\n"
		   "      if ( %s_event.func ) {\n"
		   "        typedef void (*func_type)(void *data",
	(const char *)f->flatName);
      generateCParams(out, f, GenOptions(gen_types | gen_vars | cify_types |
	gen_lead_comma));
      // finish function pointer decl start call
      fprintf(out, ");\n");
      genCCall(f, cdQtToBind, genCHookPrefix);
      fprintf(out, "      }\n"
		   "    }\n");
      first = FALSE;
    }

    // declare all private data for the events and the event handler
    first = TRUE;
    for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
      if ( first )
	fprintf(out, "  private:\n");
      fprintf(out, "    QHook %s_event;\n", (const char *)f->flatName);
      first = FALSE;
    }

    // finish declaration of the class
    fprintf(out, "};\n\n");
  }
}

QCString &generatePascalParams(BindFunction *f, QCString &result, bool add_handle,
  const char *class_name)
{
  result = "";
  int count = 0;
  BindArgument *a = f->args.first();
  if ( add_handle ) {
    result = result + "handle: " + class_name + "H";
    count++;
  }
  while ( a ) {
    if ( count++ )
      result += "; ";

    result += a->name;
    result += ": ";
    result += a->pascalType;
    a = f->args.next();
  }

  return result;
}

void generateEnumValues()
{
  const char *class_name = (const char *)className;

  QString prev_enum = "";
  for ( EnumDef *e = enumDefs.first(); e; e = enumDefs.next() ) {
    bool info_gened = FALSE;
    bool gen_value = FALSE;
    bool first = TRUE;
    const char *enum_name = e->name;

    if ( !main_gened ) {
      fprintf( out,
        "#include <stdio.h>\n"
        "#include \"%s\"\n"
        "\n"
        "#if defined(NEEDS_QMAIN)\n"
        "#undef main\n"
        "#endif\n"
        "\n"
        "int main(int, char *[]) {\n",
        (const char*)includeFile );
      main_gened = TRUE;
    }

    if ( !e->assigned )
      // This was already generated in the for Pascal types.
      // See generatePascal()
      continue;

    fprintf(out, "\tprintf(\"type\\n  %s%s = (\\n\");\n", class_name, enum_name);

    for ( EnumValue *ev = e->first(); ev; ev = e->next() ) {
      const char *name = (const char *)ev->name;
      QString ev_name = ev->name.upper();
      gen_value = gen_value || ev->value < 0;
      if ( !(ev_name == prev_enum) ) {
        if ( !first )
          fprintf(out, "\tprintf(\",\\n\");\n");
        first = FALSE;
        if ( ev->value < 0 )
          fprintf(out, "\tprintf(\"    %s%s_%s = %%d { $%%x }\", (int)%s::%s, (int)%s::%s);\n",
            class_name,
            enum_name,
            name,
            class_name,
            name,
            class_name,
            name);
        else
          fprintf(out, "\tprintf(\"    %s%s_%s\");\n",
            class_name,
            enum_name,
            name);
        info_gened = TRUE;
      }
      prev_enum = ev_name;
    }
    fprintf(out, "\tprintf(\");\\n\");\n");  
  }
  generatedCode = TRUE;
}

void generatePascal()
{
  char *hdr1 =
    "(****************************************************************************\n"
    "** %s Pascal bindings generated from reading C++ file '%s'\n**\n";
  char *hdr2 =
    "** Created: %s\n"
    "**      by: The Qt Meta Object Compiler ($Revision: 1.1 $)\n**\n";
  char *hdr3 =
    "** WARNING! All changes made in this file will be lost!\n";
  char *hdr4 =
    "*****************************************************************************)\n\n";

  int  i;
  bool destroy_emitted = FALSE;
  bool code_gened = FALSE;

  if ( skipClass && !pascalTypes )
    return;

  if ( !pascalTypes && !checkClassWarnings() )
    return;

  const char *class_name = (const char *)className;

  // generateHeader(hdr1, hdr2, hdr3, hdr4);

  if ( pascalTypes ) {
    /*
    const char *ancestor;
    QCString ancestor_s;
    // use the superclass name if the superclass was specified and it was not a template
    if ( !superclassName.isEmpty() && superclassName.find('<') < 0 ) {
      // user the first one the user typed since it require no special marshaling
      ancestor_s = *superclassList.first() + "H";
      ancestor = (const char *)ancestor_s;
    }
    else
      ancestor = "TObject";
    fprintf(out, "type\n"
	    "  %sH = class(%s) end;\n", class_name, ancestor);
    if ( vcl )
      fprintf(out, "  %s_VCLH = class(%sH) end;\n", class_name, class_name);
    fprintf(out, "\n");
    */

    // declare all supported enumerations
    for ( EnumDef *e = enumDefs.first(); e; e = enumDefs.next() ) {
      bool first = true;

      //fprintf(out, "\ntype\n"
	//	   "  %s%s = type Integer;\n", class_name, (const char *)e->name);

      if ( e->assigned )
        // assigned enums are generated through producing a C program
        // that calculates their value. See generateEnumValues().
        continue;
      fprintf(out, "\ntype\n  %s%s = ( ", class_name, (const char *)e->name);
      for ( EnumValue *ev = e->first(); ev; ev = e->next() ) {
        if ( ev->value < 0 ) break; // an assigned enum value
	if ( !first )
	  fprintf(out, ", ");
	first = false;
	fprintf(out, "%s%s_%s", class_name, (const char *)e->name,
          (const char *)ev->name);
      }
      if ( first )
        // Emit a dummy identifer so the enum is not empty
        fprintf(out, "%s%s_dummy", class_name, (const char *)e->name);
      fprintf(out, " );\n");

      code_gened = TRUE;
      generatedCode = TRUE;
    }

  }

  if ( pascalEvents ) {
    bool method_types_gened = FALSE;
    buildBindings(signals);

    for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
      QCString param_str;
      generatePascalParams(f, param_str, FALSE, class_name);
      if ( !method_types_gened )
        fprintf(out, "\ntype\n");
      fprintf(out, "  %s_%s_Event = %s (%s) of object cdecl;\n",
        class_name,
        (const char *)f->flatName,
        f->is_func ? "function" : "procedure",
        (const char *)param_str);
      method_types_gened = TRUE;
    }
    if ( method_types_gened )
      fprintf(out, "\n\n");
  }

  if ( pascalIntf ) {
    // declare the class' handle type
    if ( code_gened )
      fprintf(out, "\n");
    buildBindings(slots);

    // declare the procedures and functions
    for ( BindFunction *f = bindings.first(); f; f=bindings.next() ) {
      QCString param_str;
      generatePascalParams(f, param_str, !f->is_static && f->name != className,
			   class_name);
      char *overload_keyword = f->is_overloaded ? "overload; " : "";

      if ( f->name == className ) {
	// constructor, emit constructor and destructor
	fprintf(out, "function %s_create(%s): %sH; %scdecl;\n",
		class_name,
		(const char *)param_str,
		class_name,
		overload_keyword);

	if ( !destroy_emitted ) {
	  fprintf(out, "procedure %s_destroy(handle: %sH); cdecl;\n",
		  class_name,
		  class_name);
	  destroy_emitted = TRUE;
	}
      } else {
	fprintf(out, "%s %s_%s(%s)%s%s; %scdecl;\n",
		f->is_func ? "function" : "procedure",
		class_name,
		(const char *)f->name,
		(const char *)param_str,
		f->is_func ? ": " : "",
		f->is_func ? (const char *)f->ret_pascalType : "",
		overload_keyword);
      }
    }

    // declare conversion functions
    if ( superclassList.count() > 1 ) {
      QCString *base = superclassList.first();
      for ( base = superclassList.next(); base; base = superclassList.next() ) {
	fprintf(out, "function %s_to_%s(handle: %sH): %sH; cdecl;",
	  class_name,
	  (const char *)*base,
	  class_name,
	  (const char *)*base);
      }
    }


    // declare the event catching functions
    if ( vcl ) {
      buildBindings(signals);

      for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
        fprintf(out, "procedure %s_VCL_hook_%s(Handle: %s_VCLH; Event: %s_%s_Event); cdecl;\n",
          class_name,
          (const char *)f->flatName,
	  class_name,
          class_name,
          (const char *)f->flatName
          );
      }
    }
    fprintf(out, "\n");
  }

  if ( pascalImpl ) {
    buildBindings(slots);
    // declare implementations as external

    // declare destructor
    fprintf(out, "procedure %s_destroy; external QtShareName name QtNamePrefix + '%s_destroy';\n",
	    class_name, class_name);

    // declare additional methods and the constructor
    for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
      QCString params_s;
      char const *params = "";
      char method_num_c[10] = "";

      if ( f->is_overloaded ) {
	generatePascalParams(f, params_s, !f->is_static && f->name != className,
			     class_name);
	params_s.prepend("(");
	params_s.append(")");
	if ( f->is_func ) {
	  params_s.append(": ");
	  params_s.append(f->ret_pascalType);
	  if ( f->name == className )
	    params_s.append("H");
	}
	params = (const char *)params_s;
	if ( f->flatNumber > 1 )
	  sprintf(method_num_c, "%d", f->flatNumber);
      }
      QCString fname = f->name;
      if ( fname == className )
         fname = "create";
      fprintf(out, "%s %s_%s%s; external QtShareName name QtNamePrefix + '%s_%s%s';\n",
	      f->is_func ? "function" : "procedure",
	      class_name,
	      (const char *)fname,
	      params,
	      class_name,
	      (const char *)fname,
	      method_num_c);
    }

    // declare conversion functions
    if ( superclassList.count() > 1 ) {
      QCString *base = superclassList.first();
      for ( base = superclassList.next(); base; base = superclassList.next() ) {
	fprintf(out, "function %s_to_%s; external QtSharename name QtNamePrefix + '%s_to_%s';\n",
	  class_name,
	  (const char *)*base,
	  class_name,
	  (const char *)*base);
      }
    }
    // declare the signal hook functions
    if ( vcl ) {
      buildBindings(signals);

      for ( BindFunction *f = bindings.first(); f; f = bindings.next() ) {
	fprintf(out, "procedure %s_VCL_hook_%s; external QtShareName name QtNamePrefix + '%s_VCL_hook_%s';\n",
	  class_name,
	  (const char *)f->flatName,
	  class_name,
	  (const char *)f->flatName
	  );
      }
    }
    fprintf(out, "\n");

  }
}

void generateClass()		      // generate C++ source code for a class
{
    static int gen_count = 0;
    char *hdr1 = "/****************************************************************************\n"
		 "** %s meta object code from reading C++ file '%s'\n**\n";
    char *hdr2 = "** Created: %s\n"
		 "**      by: The Qt Meta Object Compiler ($Revision: 1.1 $)\n**\n";
    char *hdr3 = "** WARNING! All changes made in this file will be lost!\n";
    char *hdr4 = "*****************************************************************************/\n\n";
    int   i;

    if ( skipClass )				// don't generate for class
	return;
    if ( !Q_OBJECTdetected ) {
	if ( signals.count() == 0 && slots.count() == 0 )
	    return;
	generatedCode = TRUE;
	if ( displayWarnings )
	    moc_err("The declaration of the class \"%s\" contains slots "
		    "and/or signals\n\t but no Q_OBJECT macro!", className.data());
    } else {
	if ( superclassName.isEmpty() )
	    moc_err("The declaration of the class \"%s\" contains the\n"
		    "\tQ_OBJECT macro but does not inherit from any class!\n"
		    "\tInherit from QObject or one of its descendants"
		    " or remove Q_OBJECT. ", className.data() );
    }
    if ( templateClass ) {			// don't generate for class
	moc_err( "Sorry, Qt does not support templates that contain\n"
		 "signals, slots or Q_OBJECT. This will be supported soon." );
	return;
    }
    generatedCode = TRUE;
    if ( gen_count++ == 0 ) {			// first class to be generated
	QDateTime dt = QDateTime::currentDateTime();
	QCString dstr = dt.toString().ascii();
	QCString fn = fileName;
	i = fileName.length()-1;
	while ( i>0 && fileName[i-1] != '/' && fileName[i-1] != '\\' )
	    i--;				// skip path
	if ( i >= 0 )
	    fn = &fileName[i];
	fprintf( out, hdr1, (const char*)qualifiedClassName(),(const char*)fn);
	fprintf( out, hdr2, (const char*)dstr );
	fprintf( out, hdr3 );
	fprintf( out, hdr4 );
	fprintf( out, "#define Q_MOC_%s\n", className.data() );
	fprintf( out, "#if !defined(Q_MOC_OUTPUT_REVISION)\n" );
	fprintf( out, "#define Q_MOC_OUTPUT_REVISION %d\n", formatRevision );
	fprintf( out, "#elif Q_MOC_OUTPUT_REVISION != %d\n", formatRevision );
	fprintf( out, "#error \"Moc format conflict - "
		 "please regenerate all moc files\"\n" );
	fprintf( out, "#endif\n\n" );
	if ( !noInclude )
	    fprintf( out, "#include \"%s\"\n", (const char*)includeFile );
	fprintf( out, "#include <%sqmetaobject.h>\n", (const char*)qtPath );
	fprintf( out, "#include <%sqapplication.h>\n", (const char*)qtPath );
	fprintf( out, "\n" );
	fprintf( out, "#if defined(Q_SPARCWORKS_FUNCP_BUG)\n" );
	fprintf( out, "#define Q_AMPERSAND\n" );
	fprintf( out, "#else\n" );
	fprintf( out, "#define Q_AMPERSAND &\n" );
	fprintf( out, "#endif\n" );
	fprintf( out, "\n\n" );
    } else {
	fprintf( out, "\n\n" );
    }

//
// Generate virtual function className()
//
    fprintf( out, "const char *%s::className() const\n{\n    ",
	     (const char*)qualifiedClassName() );
    fprintf( out, "return \"%s\";\n}\n\n", (const char*)qualifiedClassName() );

//
// Generate static metaObj variable
//
    fprintf( out, "QMetaObject *%s::metaObj = 0;\n\n", (const char*)qualifiedClassName());

//
// Generate static meta-object constructor-object (we don't rely on
// it, except for QBuilder).
//
    fprintf( out, "\n#if QT_VERSION >= 199\n" );
    int levels = openNameSpaceForMetaObject( out );
    fprintf( out, "static QMetaObjectInit init_%s(&%s::staticMetaObject);\n\n",
	(const char*)pureClassName(), (const char*)qualifiedClassName() );
    closeNameSpaceForMetaObject( out, levels );
    fprintf( out, "#endif\n\n" );

//
// Generate initMetaObject member function
//
    fprintf( out, "void %s::initMetaObject()\n{\n", (const char*)qualifiedClassName() );
    fprintf( out, "    if ( metaObj )\n\treturn;\n" );
    fprintf( out, "    if ( strcmp(%s::className(), \"%s\") != 0 )\n"
	          "\tbadSuperclassWarning(\"%s\",\"%s\");\n",
             (const char*)qualifiedSuperclassName(), (const char*)qualifiedSuperclassName(),
             (const char*)qualifiedClassName(), (const char*)qualifiedSuperclassName() );
    fprintf( out, "\n#if QT_VERSION >= 199\n" );
    fprintf( out, "    staticMetaObject();\n");
    fprintf( out, "}\n\n");

//
// Generate tr member function
//
    fprintf( out, "QString %s::tr(const char* s)\n{\n", (const char*)qualifiedClassName() );
    fprintf( out, "    return ((QNonBaseApplication*)qApp)->translate(\"%s\",s);\n}\n\n", (const char*)qualifiedClassName() );

//
// Generate staticMetaObject member function
//
    fprintf( out, "void %s::staticMetaObject()\n{\n", (const char*)qualifiedClassName() );
    fprintf( out, "    if ( metaObj )\n\treturn;\n" );
    fprintf( out, "    %s::staticMetaObject();\n", (const char*)qualifiedSuperclassName() );
    fprintf( out, "#else\n\n" );
    fprintf( out, "    %s::initMetaObject();\n", (const char*)qualifiedSuperclassName() );
    fprintf( out, "#endif\n\n" );
//
// Build slots array in staticMetaObject()
//
    generateFuncs( &slots, "slot", Slot_Num );

//
// Build signals array in staticMetaObject()
//
    generateFuncs( &signals, "signal", Signal_Num );

//
// Finally code to create and return meta object
//
    fprintf( out, "    metaObj = QMetaObject::new_metaobject(\n"
		  "\t\"%s\", \"%s\",\n",
	     (const char*)qualifiedClassName(), (const char*)qualifiedSuperclassName() );
    if ( slots.count() )
	fprintf( out, "\tslot_tbl, %d,\n", slots.count() );
    else
	fprintf( out, "\t0, 0,\n" );
    if ( signals.count() )
	fprintf( out, "\tsignal_tbl, %d );\n", signals.count());
    else
	fprintf( out, "\t0, 0 );\n" );

    fprintf( out, "}\n" );

//
// End of function initMetaObject()
//

//
// Generate internal signal functions
//
    Function *f;
    f = signals.first();			// make internal signal methods
    static bool included_list_stuff = FALSE;
    while ( f ) {
	QCString typstr = "";			// type string
	QCString valstr = "";			// value string
	QCString argstr = "";			// argument string (type+value)
	char	buf[12];
	Argument *a = f->args->first();
	QCString typvec[32], valvec[32], argvec[32];
	typvec[0] = "";
	valvec[0] = "";
	argvec[0] = "";

	i = 0;
	while ( a ) {				// argument list
	    if ( !a->leftType.isEmpty() || !a->rightType.isEmpty() ) {
		if ( i ) {
		    typstr += ",";
		    valstr += ", ";
		    argstr += ", ";
		}
		typstr += a->leftType;
		typstr += a->rightType;
		argstr += a->leftType;
		argstr += " ";
		sprintf( buf, "t%d", i );
		valstr += buf;
		argstr += buf;
		argstr += a->rightType;
		++i;
		typvec[i] = typstr.copy();
		valvec[i] = valstr.copy();
		argvec[i] = argstr.copy();
	    }
	    a = f->args->next();
	}

	const char *predef_call_func;
	if ( typstr == "bool" ) {
	    predef_call_func = "activate_signal_bool";
	} else if ( typstr == "QString" ) {
	    predef_call_func = "activate_signal_string";
	} else if ( typstr == "const QString&" ) {
	    predef_call_func = "activate_signal_strref";
	} else if ( typstr.isEmpty() || typstr == "short" ||
		    typstr == "int" ||  typstr == "long" ||
		    typstr == "char*" || typstr == "const char*" ) {
	    predef_call_func = "activate_signal";
	} else {
	    predef_call_func = 0;
	}
	if ( !predef_call_func && !included_list_stuff ) {
	    // yes we need it, because otherwise QT_VERSION may not be defined
	    fprintf( out, "\n#include <%sqobjectdefs.h>\n", (const char*)qtPath );
	    fprintf( out, "#include <%sqsignalslotimp.h>\n", (const char*)qtPath );
	    included_list_stuff = TRUE;
	}

	fprintf( out, "\n/" /* c++ */ "/ SIGNAL %s\n", (const char*)f->name );
	fprintf( out, "void %s::%s(", (const char*)qualifiedClassName(),
		 (const char*)f->name );

	if ( argstr.isEmpty() )
	    fprintf( out, ")\n{\n" );
	else
	    fprintf( out, " %s )\n{\n", (const char*)argstr );

	if ( predef_call_func ) {
	    fprintf( out, "    %s( \"%s(%s)\"", predef_call_func,
		     (const char*)f->name, (const char*)typstr );
	    if ( !valstr.isEmpty() )
		fprintf( out, ", %s", (const char*)valstr );
	    fprintf( out, " );\n}\n" );
	    f = signals.next();
	    continue;
	}


	fprintf( out,"    // No builtin function for signal parameter type %s\n",
		 (const char*)typstr );
	int nargs = f->args->count();
	fprintf( out, "    QConnectionList *clist = receivers(\"%s(%s)\");\n",
		 (const char*)f->name, (const char*)typstr );
	fprintf( out, "    if ( !clist || signalsBlocked() )\n\treturn;\n" );
	if ( nargs ) {
	    for ( i=0; i<=nargs; i++ ) {
		fprintf( out, "    typedef void (QObject::*RT%d)(%s);\n",
			 i, (const char*)typvec[i] );
		fprintf( out, "    typedef RT%d *PRT%d;\n", i, i );
	    }
	} else {
	    fprintf( out, "    typedef void (QObject::*RT)(%s);\n",
		     (const char*)typstr);
	    fprintf( out, "    typedef RT *PRT;\n" );
	}
	if ( nargs ) {
	    for ( i=0; i<=nargs; i++ )
		fprintf( out, "    RT%d r%d;\n", i, i );
	} else {
	    fprintf( out, "    RT r;\n" );
	}
	fprintf( out, "    QConnectionListIt it(*clist);\n" );
	fprintf( out, "    QConnection   *c;\n" );
	fprintf( out, "    QSenderObject *object;\n" );
	fprintf( out, "    while ( (c=it.current()) ) {\n" );
	fprintf( out, "\t++it;\n" );
	fprintf( out, "\tobject = (QSenderObject*)c->object();\n" );
	fprintf( out, "\tobject->setSender( this );\n" );
	if ( nargs ) {
	    fprintf( out, "\tswitch ( c->numArgs() ) {\n" );
	    for ( i=0; i<=nargs; i++ ) {
		fprintf( out, "\t    case %d:\n", i );
		fprintf( out, "\t\tr%d = *((PRT%d)(c->member()));\n", i, i );
		fprintf( out, "\t\t(object->*r%d)(%s);\n",
			 i, (const char*)valvec[i] );
		fprintf( out, "\t\tbreak;\n" );
	    }
	    fprintf( out, "\t}\n" );
	} else {
	    fprintf( out, "\tr = *((PRT)(c->member()));\n" );
	    fprintf( out, "\t(object->*r)(%s);\n", (const char*)valstr );
	}
	fprintf( out, "    }\n}\n" );
	f = signals.next();
    }

}


ArgList *addArg( Argument *a )			// add argument to list
{
    tmpArgList->append( a );
    return tmpArgList;
}

EnumDef *addEnum( EnumValue *e)
{
  if ( tmpEnum && e ) {
    tmpEnum->append(e);
    return tmpEnum;
  }
  return NULL;
}

void addEnumDef( EnumDef *e )
{
  if ( e )
    enumDefs.append(e);
}

void addMember( char m )
{
    if ( skipFunc ) {
	tmpArgList  = new ArgList;   // ugly but works!
	CHECK_PTR( tmpArgList );
	tmpFunc	    = new Function;
	CHECK_PTR( tmpFunc );
	skipFunc    = FALSE;
	staticFunc  = FALSE;
	return;
    }
    if ( m == 's' && tmpFunc->type != "void" ) {
	moc_err( "Signals must have \"void\" as their return type" );
	return;
    }
    tmpFunc->accessPerm = tmpAccessPerm;
    tmpFunc->args	= tmpArgList;
    tmpFunc->lineNo	= lineNo;
    tmpFunc->is_static	= staticFunc;
    tmpArgList		= new ArgList;
    CHECK_PTR( tmpArgList );
    switch( m ) {
	case 's': signals.append( tmpFunc ); break;
	case 't': slots.  append( tmpFunc ); break;
    }
    skipFunc = FALSE;
    staticFunc = FALSE;
    tmpFunc  = new Function;
    CHECK_PTR( tmpFunc );
}
